from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.permissions import (
    permission_document_file_new, permission_document_file_view
)
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..events import (
    event_document_checked_in, event_document_checked_out,
    event_document_forcefully_checked_in
)
from ..exceptions import DocumentAlreadyCheckedOut, DocumentNotCheckedOut
from ..literals import STATE_CHECKED_OUT, STATE_LABELS
from ..permissions import (
    permission_document_check_in, permission_document_check_in_override,
    permission_document_check_out, permission_document_check_out_detail_view
)

from .mixins import DocumentCheckoutTestMixin, DocumentCheckoutViewTestMixin


class DocumentCheckoutViewTestCase(
    DocumentCheckoutTestMixin, DocumentCheckoutViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_document_check_in_get_view_no_permission(self):
        self._check_out_test_document()

        self._clear_events()

        response = self._request_test_document_check_in_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_check_in_get_view_with_access(self):
        self._check_out_test_document()

        self.grant_access(
            obj=self._test_document, permission=permission_document_check_in
        )

        self._clear_events()

        response = self._request_test_document_check_in_get_view()
        self.assertContains(
            response=response, text=self._test_document.label,
            status_code=200
        )

        self.assertTrue(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_check_in_get_view_with_access(self):
        self._check_out_test_document()

        self.grant_access(
            obj=self._test_document, permission=permission_document_check_in
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_check_in_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_check_in_post_view_no_permission(self):
        self._check_out_test_document()

        self._clear_events()

        response = self._request_test_document_check_in_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_check_in_post_view_with_access(self):
        self._check_out_test_document()

        self.grant_access(
            obj=self._test_document, permission=permission_document_check_in
        )

        self._clear_events()

        response = self._request_test_document_check_in_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertFalse(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_checked_in.id)

    def test_trashed_document_check_in_post_view_with_access(self):
        self._check_out_test_document()

        self.grant_access(
            obj=self._test_document, permission=permission_document_check_in
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_check_in_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_check_in_post_view_no_permission(self):
        # Upload second document.
        self._upload_test_document()

        self._check_out_test_document(
            document=self._test_documents[0]
        )
        self._check_out_test_document(
            document=self._test_documents[1]
        )

        self._clear_events()

        response = self._request_test_document_multiple_check_in_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_documents[0].is_checked_out()
        )
        self.assertTrue(
            self._test_documents[1].is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_check_in_post_view_with_document_0_access(self):
        # Upload second document.
        self._upload_test_document()

        self._check_out_test_document(
            document=self._test_documents[0]
        )
        self._check_out_test_document(
            document=self._test_documents[1]
        )

        self.grant_access(
            obj=self._test_documents[0],
            permission=permission_document_check_in
        )

        self._clear_events()

        response = self._request_test_document_multiple_check_in_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertFalse(
            self._test_documents[0].is_checked_out()
        )
        self.assertTrue(
            self._test_documents[1].is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_documents[0])
        self.assertEqual(events[0].verb, event_document_checked_in.id)

    def test_document_multiple_check_in_post_view_with_access(self):
        # Upload second document.
        self._upload_test_document()

        self._check_out_test_document(
            document=self._test_documents[0]
        )
        self._check_out_test_document(
            document=self._test_documents[1]
        )

        self.grant_access(
            obj=self._test_documents[0],
            permission=permission_document_check_in
        )
        self.grant_access(
            obj=self._test_documents[1],
            permission=permission_document_check_in
        )

        self._clear_events()

        response = self._request_test_document_multiple_check_in_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertFalse(
            self._test_documents[0].is_checked_out()
        )
        self.assertFalse(
            self._test_documents[1].is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_documents[0])
        self.assertEqual(events[0].verb, event_document_checked_in.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_documents[1])
        self.assertEqual(events[1].verb, event_document_checked_in.id)

    def test_document_check_out_get_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_check_out_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertFalse(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_check_out_get_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_check_out
        )

        self._clear_events()

        response = self._request_test_document_check_out_get_view()
        self.assertEqual(response.status_code, 200)

        self.assertFalse(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_check_out_get_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_check_out
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_check_out_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertFalse(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_check_out_post_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_check_out_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertFalse(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_check_out_post_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_check_out
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_check_out_detail_view
        )

        self._clear_events()

        response = self._request_test_document_check_out_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_checked_out.id)

    def test_trashed_document_check_out_post_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_check_out
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_check_out_detail_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_check_out_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertFalse(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_check_out_post_view_no_permission(self):
        # Upload second document.
        self._upload_test_document()

        self.grant_access(
            obj=self._test_documents[0],
            permission=permission_document_check_out_detail_view
        )
        self.grant_access(
            obj=self._test_documents[1],
            permission=permission_document_check_out_detail_view
        )

        self._clear_events()

        response = self._request_test_document_multiple_check_out_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertFalse(
            self._test_documents[0].is_checked_out()
        )
        self.assertFalse(
            self._test_documents[1].is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_check_out_post_view_with_document_access(self):
        # Upload second document.
        self._upload_test_document()

        self.grant_access(
            obj=self._test_documents[0],
            permission=permission_document_check_out
        )
        self.grant_access(
            obj=self._test_documents[0],
            permission=permission_document_check_out_detail_view
        )
        self.grant_access(
            obj=self._test_documents[1],
            permission=permission_document_check_out_detail_view
        )

        self._clear_events()

        response = self._request_test_document_multiple_check_out_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_documents[0].is_checked_out()
        )
        self.assertFalse(
            self._test_documents[1].is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_documents[0])
        self.assertEqual(events[0].verb, event_document_checked_out.id)

    def test_document_multiple_check_out_post_view_with_access(self):
        # Upload second document.
        self._upload_test_document()

        self.grant_access(
            obj=self._test_documents[0],
            permission=permission_document_check_out
        )
        self.grant_access(
            obj=self._test_documents[1],
            permission=permission_document_check_out
        )
        self.grant_access(
            obj=self._test_documents[0],
            permission=permission_document_check_out_detail_view
        )
        self.grant_access(
            obj=self._test_documents[1],
            permission=permission_document_check_out_detail_view
        )

        self._clear_events()

        response = self._request_test_document_multiple_check_out_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_documents[0].is_checked_out()
        )
        self.assertTrue(
            self._test_documents[1].is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_documents[0])
        self.assertEqual(events[0].verb, event_document_checked_out.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_documents[1])
        self.assertEqual(events[1].verb, event_document_checked_out.id)

    def test_document_check_out_detail_view_no_permission(self):
        self._check_out_test_document()

        self._clear_events()

        response = self._request_test_document_check_out_detail_view()

        self.assertNotContains(
            response, text=STATE_LABELS[STATE_CHECKED_OUT], status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_check_out_detail_view_with_access(self):
        self._check_out_test_document()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_check_out_detail_view
        )

        self._clear_events()

        response = self._request_test_document_check_out_detail_view()
        self.assertContains(
            response, text=STATE_LABELS[STATE_CHECKED_OUT], status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_check_out_detail_view_with_access(self):
        self._check_out_test_document()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_check_out_detail_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_check_out_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_check_out_list_view_no_permission(self):
        self._check_out_test_document()

        self._clear_events()

        response = self._request_test_document_check_out_list_view()
        self.assertNotContains(
            response=response, text=self._test_document.label,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_check_out_list_view_with_access(self):
        self._check_out_test_document()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_check_out_detail_view
        )

        self._clear_events()

        response = self._request_test_document_check_out_list_view()
        self.assertContains(
            response=response, text=self._test_document.label,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_check_out_list_view_with_access(self):
        self._check_out_test_document()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_check_out_detail_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_check_out_list_view()
        self.assertNotContains(
            response=response, text=self._test_document.label,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_check_in_forcefull_view_no_permission(self):
        # Gitlab issue #237
        # Forcefully checking in a document by a user without adequate
        # permissions throws out an error.
        self._create_test_user()
        # Check out document as test_user.
        self._check_out_test_document(user=self._test_user)

        self.grant_access(
            obj=self._test_document, permission=permission_document_check_in
        )

        self._clear_events()

        response = self._request_test_document_check_in_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_check_in_forcefull_view_with_access(self):
        self._create_test_user()
        # Check out document as test_user.
        self._check_out_test_document(user=self._test_user)

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_check_in_override
        )

        self._clear_events()

        # Check in document as test_case_user.
        response = self._request_test_document_check_in_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertFalse(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(
            events[0].verb, event_document_forcefully_checked_in.id
        )

    def test_trashed_document_check_in_forcefull_view_with_access(self):
        self._create_test_user()
        # Check out document as test_user
        self._check_out_test_document(user=self._test_user)

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_check_in_override
        )

        self._test_document.delete()

        self._clear_events()

        # Check in document as test_case_user
        response = self._request_test_document_check_in_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_check_in_of_non_checked_out_document_view(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_check_in
        )

        self._clear_events()

        response = self._request_test_document_check_in_post_view(
            follow=True
        )
        self.assertContains(
            response=response, status_code=200,
            text=str(
                DocumentNotCheckedOut()
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_check_in_of_non_checked_out_trashed_document_view(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_check_in
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_check_in_post_view(
            follow=True
        )
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_check_out_of_checked_out_document_view(self):
        self._create_test_user()
        self._check_out_test_document(user=self._test_user)
        self.grant_access(
            obj=self._test_document, permission=permission_document_check_out
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_check_out_detail_view
        )

        self._clear_events()

        response = self._request_test_document_check_out_post_view(
            follow=True
        )
        self.assertContains(
            response=response, status_code=200,
            text=str(
                DocumentAlreadyCheckedOut()
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_check_out_of_checked_out_trashed_document(self):
        self._create_test_user()
        self._check_out_test_document(user=self._test_user)
        self.grant_access(
            obj=self._test_document, permission=permission_document_check_out
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_check_out_detail_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_check_out_post_view(
            follow=True
        )
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentCheckoutOptionsViewTestCase(
    DocumentCheckoutTestMixin, DocumentCheckoutViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_document_check_out_block_new_file_view(self):
        self._silence_logger(name='mayan.apps.sources.links')

        self._create_test_user()
        self._check_out_test_document(user=self._test_user)
        file_count = DocumentFile.objects.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_new
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_view
        )

        self._clear_events()

        response = self.post(
            viewname='sources:document_file_upload', kwargs={
                'document_id': self._test_document.pk
            }, follow=True
        )
        self.assertContains(
            response=response, status_code=200,
            text='Unable to upload new files'
        )

        self.assertEqual(DocumentFile.objects.count(), file_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_check_out_block_new_version(self):
        self._check_out_test_document()
        file_count = DocumentFile.objects.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_new
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self.post(
            viewname='sources:document_file_upload', kwargs={
                'document_id': self._test_document.pk
            }, follow=True
        )
        self.assertEqual(response.status_code, 404)

        self.assertEqual(DocumentFile.objects.count(), file_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from contextlib import contextmanager
import copy
import os

from django.conf import settings
from django.utils.encoding import force_bytes

from mayan.apps.storage.utils import NamedTemporaryFile, fs_cleanup
from mayan.apps.testing.tests.mixins import EnvironmentTestCaseMixin

from ..classes import Setting, SettingNamespace
from ..utils import BaseSetting, SettingNamespaceSingleton

from .literals import (
    TEST_BOOTSTAP_SETTING_NAME, TEST_NAMESPACE_LABEL, TEST_NAMESPACE_NAME,
    TEST_SETTING_DEFAULT_VALUE, TEST_SETTING_GLOBAL_NAME
)


class BoostrapSettingTestMixin:
    def _create_test_bootstrap_singleton(self):
        self.test_globals = {}
        self.test_globals['BASE_DIR'] = ''
        self.test_setting_namespace_singleton = SettingNamespaceSingleton(
            global_symbol_table=self.test_globals
        )

    def _register_test_boostrap_setting(self):
        SettingNamespaceSingleton.register_setting(
            name=TEST_BOOTSTAP_SETTING_NAME, klass=BaseSetting, kwargs={
                'has_default': True, 'default_value': 'value default'
            }
        )


class SmartSettingsTestCaseMixin:
    def setUp(self):
        super().setUp()
        SettingNamespace.invalidate_cache_all()

        with NamedTemporaryFile(delete=False) as self.test_setting_config_file_object:
            settings.CONFIGURATION_FILEPATH = self.test_setting_config_file_object.name
            os.environ['MAYAN_CONFIGURATION_FILEPATH'] = self.test_setting_config_file_object.name
            Setting._config_file_cache = None

    def tearDown(self):
        fs_cleanup(filename=self.test_setting_config_file_object.name)
        SettingNamespace.invalidate_cache_all()
        super().tearDown()

    @contextmanager
    def override_setting(self, global_name, method=None, value=None):
        setting = Setting.get(global_name=global_name)
        old_value = copy.copy(setting.value)
        if method is None:
            setting.set(value=value)
        else:
            current_value = setting.value
            getattr(current_value, method)(value)
            setting.set(value=current_value)
        try:
            yield
        finally:
            setting.set(value=old_value)


class SmartSettingNamespaceTestMixin(EnvironmentTestCaseMixin):
    def _create_test_settings_namespace(self, **kwargs):
        try:
            self.test_settings_namespace = SettingNamespace.get(
                name=TEST_NAMESPACE_NAME
            )
            self.test_settings_namespace.migration_class = None
            self.test_settings_namespace.version = None
            self.test_settings_namespace.__dict__.update(kwargs)
        except KeyError:
            self.test_settings_namespace = SettingNamespace(
                label=TEST_NAMESPACE_LABEL, name=TEST_NAMESPACE_NAME,
                **kwargs
            )


class SmartSettingNamespaceViewTestMixin:
    def _request_namespace_list_view(self):
        return self.get(viewname='settings:setting_namespace_list')

    def _request_namespace_detail_view(self):
        return self.get(
            viewname='settings:setting_namespace_detail', kwargs={
                'namespace_name': self.test_settings_namespace.name
            }
        )


class SmartSettingTestMixin(EnvironmentTestCaseMixin):
    test_setting_global_name = None
    test_config_file_object = None

    def tearDown(self):
        if self.test_config_file_object:
            fs_cleanup(filename=self.test_config_file_object.name)
        super().tearDown()

    def _create_test_config_file(self, callback=None):
        if not self.test_setting_global_name:
            self.test_setting_global_name = self.test_setting.global_name

        test_config_entry = {
            self.test_setting_global_name: self.test_config_value
        }

        with NamedTemporaryFile(delete=False) as test_config_file_object:
            # Needed to load the config file from the Setting class
            # after bootstrap.
            settings.CONFIGURATION_FILEPATH = test_config_file_object.name
            # Needed to update the globals before Mayan has loaded.
            self._set_environment_variable(
                name='MAYAN_CONFIGURATION_FILEPATH',
                value=test_config_file_object.name
            )
            test_config_file_object.write(
                force_bytes(
                    s=Setting.serialize_value(value=test_config_entry)
                )
            )
            test_config_file_object.seek(0)
            Setting._config_file_cache = None

            if callback:
                callback()

    def _create_test_setting(self, validation_function=None):
        self.test_setting = self.test_settings_namespace.add_setting(
            global_name=TEST_SETTING_GLOBAL_NAME,
            default=TEST_SETTING_DEFAULT_VALUE,
            validation_function=validation_function
        )


class SmartSettingViewTestMixin:
    def _request_setting_edit_view(self, value):
        return self.post(
            viewname='settings:setting_edit_view', kwargs={
                'setting_global_name': self.test_setting.global_name
            }, data={'value': value}
        )



from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.documents.tests.mixins.document_version_mixins import DocumentVersionLinkTestMixin

from ..links import link_document_version_ocr_content_single_delete
from ..permissions import permission_document_version_ocr


class DocumentOCRLinkTestCase(
    DocumentVersionLinkTestMixin, GenericDocumentViewTestCase
):
    def test_document_version_ocr_content_single_delete_link_no_permission(self):
        resolved_link = self._resolve_test_document_version_link(
            test_link=link_document_version_ocr_content_single_delete
        )
        self.assertEqual(resolved_link, None)

    def test_document_version_ocr_content_single_delete_link_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_version_ocr
        )
        resolved_link = self._resolve_test_document_version_link(
            test_link=link_document_version_ocr_content_single_delete
        )
        self.assertNotEqual(resolved_link, None)



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0059_auto_20200918_0616')
    ]
    operations = [
        migrations.CreateModel(
            name='DocumentVersion',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'timestamp', models.DateTimeField(
                        auto_now_add=True, db_index=True,
                        help_text='The server date and time when the '
                        'document version was created.',
                        verbose_name='Timestamp'
                    )
                ),
                (
                    'comment', models.TextField(
                        blank=True, default='',
                        help_text='An optional short text describing '
                        'the document version.', verbose_name='Comment'
                    )
                ),
                (
                    'document', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='versions', to='documents.Document',
                        verbose_name='Document'
                    )
                )
            ],
            options={
                'verbose_name': 'Document version',
                'verbose_name_plural': 'Document versions',
                'ordering': ('timestamp',)
            }
        )
    ]



from .events import event_role_edited


def method_group_roles_add(self, queryset, user=None):
    for role in queryset:
        self.roles.add(role)
        event_role_edited.commit(
            action_object=self, actor=user, target=role
        )


def method_group_roles_remove(self, queryset, user=None):
    for role in queryset:
        self.roles.remove(role)
        event_role_edited.commit(
            action_object=self, actor=user, target=role
        )



import hashlib
import logging

from furl import furl
from graphviz import Digraph

from django.apps import apps
from django.core import serializers
from django.db import IntegrityError
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from mayan.apps.file_caching.models import CachePartitionFile

from ..events import event_workflow_template_edited
from ..literals import (
    STORAGE_NAME_WORKFLOW_CACHE, SYMBOL_GRAPHVIZ_CONDITIONAL,
    WORKFLOW_ACTION_ON_ENTRY
)

logger = logging.getLogger(name=__name__)


class WorkflowBusinessLogicMixin:
    @cached_property
    def cache(self):
        Cache = apps.get_model(app_label='file_caching', model_name='Cache')

        return Cache.objects.get(
            defined_storage_name=STORAGE_NAME_WORKFLOW_CACHE
        )

    @cached_property
    def cache_partition(self):
        partition, created = self.cache.partitions.get_or_create(
            name='{}'.format(self.pk)
        )
        return partition

    def document_types_add(self, queryset, user):
        for model_instance in queryset.all():
            self.document_types.add(model_instance)
            event_workflow_template_edited.commit(
                action_object=model_instance, actor=user, target=self
            )

    def document_types_remove(self, queryset, user):
        for model_instance in queryset.all():
            self.document_types.remove(model_instance)
            event_workflow_template_edited.commit(
                action_object=model_instance, actor=user, target=self
            )
            self.instances.filter(
                document__document_type_id=model_instance.pk
            ).delete()

    def generate_image(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        # `user` argument added for compatibility.
        cache_filename = '{}'.format(
            self.get_hash()
        )

        try:
            self.cache_partition.get_file(filename=cache_filename)
        except CachePartitionFile.DoesNotExist:
            logger.debug(
                'workflow cache file "%s" not found', cache_filename
            )

            image = self.render()
            with self.cache_partition.create_file(filename=cache_filename) as file_object:
                file_object.write(image)
        else:
            logger.debug(
                'workflow cache file "%s" found', cache_filename
            )

        return cache_filename

    def get_api_image_url(self, *args, **kwargs):
        final_url = furl()
        final_url.args = kwargs
        final_url.path = reverse(
            viewname='rest_api:workflow-template-image',
            kwargs={'workflow_template_id': self.pk}
        )
        final_url.args['_hash'] = self.get_hash()

        return final_url.tostr()

    def get_document_types_not_in_workflow(self):
        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )

        return DocumentType.objects.exclude(
            pk__in=self.document_types.all()
        )

    def get_hash(self):
        result = hashlib.sha256(
            string=serializers.serialize(
                format='json', queryset=(self,)
            ).encode()
        )
        for state in self.states.all():
            result.update(
                state.get_hash().encode()
            )

        for transition in self.transitions.all():
            result.update(
                transition.get_hash().encode()
            )

        return result.hexdigest()

    def get_initial_state(self):
        try:
            return self.states.get(initial=True)
        except self.states.model.DoesNotExist:
            return None
    get_initial_state.short_description = _('Initial state')

    def launch_for(self, document, user=None):
        if document.document_type in self.document_types.all():
            try:
                logger.info(
                    'Launching workflow %s for document %s', self, document
                )
                WorkflowInstance = apps.get_model(
                    app_label='document_states',
                    model_name='WorkflowInstance'
                )
                workflow_instance = WorkflowInstance(
                    document=document, workflow=self
                )
                workflow_instance._event_actor = user
                workflow_instance.save()

                initial_state = self.get_initial_state()
                if initial_state:
                    for action in initial_state.entry_actions.filter(enabled=True):
                        context = workflow_instance.get_context()
                        context.update(
                            {
                                'action': action
                            }
                        )
                        action.execute(
                            context=context,
                            workflow_instance=workflow_instance
                        )
            except IntegrityError:
                logger.info(
                    'Workflow %s already launched for document %s',
                    self, document
                )
            else:
                logger.info(
                    'Workflow %s launched for document %s', self, document
                )
                return workflow_instance
        else:
            logger.error(
                'This workflow is not valid for the document type of the '
                'document.'
            )

    def render(self):
        diagram = Digraph(
            name='finite_state_machine', graph_attr={
                'rankdir': 'LR', 'splines': 'polyline'
            }, format='png'
        )

        action_cache = {}
        escalation_cache = {}
        state_cache = {}
        transition_cache = []

        for state in self.states.all():
            state_cache[
                's{}'.format(state.pk)
            ] = {
                'connections': {'origin': 0, 'destination': 0},
                'initial': state.initial,
                'label': state.label,
                'name': 's{}'.format(state.pk)
            }

            for action in state.actions.all():
                if action.enabled:
                    if action.has_condition():
                        action_label = '{} {}'.format(
                            SYMBOL_GRAPHVIZ_CONDITIONAL, action.label
                        )
                    else:
                        action_label = action.label

                    action_cache[
                        'a{}'.format(action.pk)
                    ] = {
                        'label': action_label,
                        'name': 'a{}'.format(action.pk),
                        'state': 's{}'.format(state.pk),
                        'when': action.when
                    }

            for escalation in state.escalations.all():
                if escalation.enabled:
                    label = 'After {}'.format(
                        escalation.get_time_display()
                    )

                    if escalation.has_condition():
                        label = '{} {}'.format(
                            SYMBOL_GRAPHVIZ_CONDITIONAL, label
                        )

                    escalation_cache[
                        'e{}'.format(escalation.pk)
                    ] = {
                        'label': label,
                        'destination': 's{}'.format(escalation.transition.destination_state.pk),
                        'state': 's{}'.format(state.pk)
                    }

        for transition in self.transitions.all():
            if transition.has_condition():
                transition_label = '{} {}'.format(
                    SYMBOL_GRAPHVIZ_CONDITIONAL, transition.label
                )
            else:
                transition_label = transition.label

            transition_cache.append(
                {
                    'label': transition_label,
                    'head_name': 's{}'.format(
                        transition.destination_state.pk
                    ),
                    'tail_name': 's{}'.format(transition.origin_state.pk)
                }
            )

            state_cache[
                's{}'.format(transition.origin_state.pk)
            ]['connections']['origin'] = state_cache[
                's{}'.format(transition.origin_state.pk)
            ]['connections']['origin'] + 1

            state_cache[
                's{}'.format(transition.destination_state.pk)
            ]['connections']['destination'] += 1

        for key, value in state_cache.items():
            kwargs = {
                'fillcolor': '#eeeeee',
                'label': value['label'],
                'name': value['name'],
                'shape': 'doublecircle' if value['connections']['origin'] == 0 or value['connections']['destination'] == 0 or value['initial'] else 'circle',
                'style': 'filled' if value['initial'] else ''
            }
            diagram.node(**kwargs)

        for transition in transition_cache:
            diagram.edge(**transition)

        for key, value in action_cache.items():
            kwargs = {
                'label': value['label'],
                'name': value['name'],
                'shape': 'box'
            }
            diagram.node(**kwargs)
            diagram.edge(
                **{
                    'arrowhead': 'dot',
                    'arrowtail': 'dot',
                    'dir': 'both',
                    'label': 'On entry' if value['when'] == WORKFLOW_ACTION_ON_ENTRY else 'On exit',
                    'head_name': '{}'.format(
                        value['name']
                    ),
                    'style': 'dashed',
                    'tail_name': '{}'.format(
                        value['state']
                    )
                }
            )

        for key, value in escalation_cache.items():
            diagram.edge(
                **{
                    'label': value['label'],
                    'head_name': '{}'.format(
                        value['destination']
                    ),
                    'style': 'dashed',
                    'tail_name': '{}'.format(
                        value['state']
                    )
                }
            )

        return diagram.pipe()



from mayan.apps.events.classes import EventType
from mayan.apps.events.tests.mixins import EventTypeTestMixin

from ...models.workflow_transition_models import WorkflowTransitionTriggerEvent

from .workflow_template_transition_mixins import WorkflowTemplateTransitionTestMixin


class WorkflowTemplateTransitionTriggerTestMixin(
    EventTypeTestMixin, WorkflowTemplateTransitionTestMixin
):
    _test_object_model = WorkflowTransitionTriggerEvent
    _test_object_name = '_test_workflow_template_transition_trigger'

    def setUp(self):
        super().setUp()
        self._test_workflow_template_transition_triggers = []

    def _create_test_workflow_template_transition_trigger(self):
        event_type = EventType.get(id=self._test_event_type.id)

        self._test_workflow_template_transition_trigger = self._test_workflow_template_transition.trigger_events.create(
            event_type=event_type.get_stored_event_type()
        )
        self._test_workflow_template_transition_triggers.append(
            self._test_workflow_template_transition_trigger
        )


class WorkflowTemplateTransitionTriggerAPIViewTestMixin(
    WorkflowTemplateTransitionTriggerTestMixin
):
    def _request_test_workflow_template_transition_trigger_create_api_view(self):
        data = {
            'event_type_id': self._test_event_type.id
        }

        self._test_object_track()

        response = self.post(
            viewname='rest_api:workflow-template-transition-trigger-list',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }, data=data
        )

        self._test_object_set()

        return response

    def _request_test_workflow_template_transition_trigger_delete_api_view(self):
        return self.delete(
            viewname='rest_api:workflow-template-transition-trigger-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk,
                'workflow_template_transition_trigger_id': self._test_workflow_template_transition_trigger.pk
            }
        )

    def _request_test_workflow_template_transition_trigger_detail_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-trigger-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk,
                'workflow_template_transition_trigger_id': self._test_workflow_template_transition_trigger.pk
            }
        )

    def _request_test_workflow_template_transition_trigger_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-trigger-list',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }
        )

    def _request_test_workflow_template_transition_trigger_edit_patch_api_view(self):
        return self.patch(
            viewname='rest_api:workflow-template-transition-trigger-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk,
                'workflow_template_transition_trigger_id': self._test_workflow_template_transition_trigger.pk
            }, data={
                'event_type_id': self._test_event_type.id
            }
        )

    def _request_test_workflow_template_transition_trigger_edit_put_api_view(self):
        return self.put(
            viewname='rest_api:workflow-template-transition-trigger-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk,
                'workflow_template_transition_trigger_id': self._test_workflow_template_transition_trigger.pk
            }, data={
                'event_type_id': self._test_event_type.id
            }
        )


class WorkflowTemplateTransitionTriggerViewTestMixin(
    WorkflowTemplateTransitionTriggerTestMixin
):
    def _request_test_workflow_template_transition_event_list_view(self):
        return self.get(
            viewname='document_states:workflow_template_transition_triggers',
            kwargs={
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }
        )



from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('checkouts', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='documentcheckout',
            name='user',
            field=models.ForeignKey(
                blank=True, null=True, on_delete=models.CASCADE,
                to=settings.AUTH_USER_MODEL, verbose_name='User'
            ),
            preserve_default=True,
        ),
    ]



from mayan.apps.rest_api import generics

from ..models import Source
from ..permissions import (
    permission_sources_create, permission_sources_delete,
    permission_sources_edit, permission_sources_view
)
from ..serializers import SourceSerializer


class APISourceListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the source.
    post: Create a new source.
    """
    mayan_object_permissions = {
        'GET': (permission_sources_view,)
    }
    mayan_view_permissions = {
        'POST': (permission_sources_create,)
    }
    serializer_class = SourceSerializer
    source_queryset = Source.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APISourceView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected source.
    get: Details of the selected source.
    patch: Edit the selected source.
    put: Edit the selected source.
    """
    lookup_url_kwarg = 'source_id'
    mayan_object_permissions = {
        'DELETE': (permission_sources_delete,),
        'GET': (permission_sources_view,),
        'PATCH': (permission_sources_edit,),
        'PUT': (permission_sources_edit,)
    }
    serializer_class = SourceSerializer
    source_queryset = Source.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_document_index_instance_list, icon_document_type_index_templates,
    icon_index, icon_index_instances_rebuild, icon_index_instances_reset,
    icon_index_template_create, icon_index_template_delete,
    icon_index_template_document_types, icon_index_template_edit,
    icon_index_template_event_triggers, icon_index_template_list,
    icon_index_template_node_create, icon_index_template_node_delete,
    icon_index_template_node_edit, icon_index_template_node_tree_view
)
from .permissions import (
    permission_index_instance_view, permission_index_template_create,
    permission_index_template_delete, permission_index_template_edit,
    permission_index_template_rebuild, permission_index_template_view
)


def condition_is_not_root_node(context, resolved_object):
    return not resolved_object.is_root_node()


# Document type

link_document_index_instance_list = Link(
    args='resolved_object.pk', icon=icon_document_index_instance_list,
    permissions=(permission_index_instance_view,),
    text=_('Indexes'), view='indexing:document_index_list'
)

link_document_type_index_templates = Link(
    args='resolved_object.pk', icon=icon_document_type_index_templates,
    permissions=(permission_index_template_create,),
    text=_('Index templates'), view='indexing:document_type_index_templates'
)

# Index instance

link_index_instance_menu = Link(
    condition=factory_condition_queryset_access(
        app_label='document_indexing', model_name='IndexTemplate',
        object_permission=permission_index_instance_view,
    ), icon=icon_index,
    text=_('Indexes'), view='indexing:index_list'
)
link_index_instance_rebuild = Link(
    args='resolved_object.pk', icon=icon_index_instances_rebuild,
    permissions=(permission_index_template_rebuild,),
    text=_('Rebuild index'), view='indexing:index_template_rebuild'
)
link_index_instances_rebuild = Link(
    condition=factory_condition_queryset_access(
        app_label='document_indexing', model_name='IndexTemplate',
        object_permission=permission_index_template_rebuild,
    ), description=_(
        'Deletes and creates from scratch all the document indexes.'
    ), icon=icon_index_instances_rebuild, text=_('Rebuild indexes'),
    view='indexing:rebuild_index_instances'
)
link_index_instances_reset = Link(
    condition=factory_condition_queryset_access(
        app_label='document_indexing', model_name='IndexTemplate',
        object_permission=permission_index_template_rebuild,
    ), description=_(
        'Deletes and creates from scratch all the document indexes.'
    ), icon=icon_index_instances_reset, text=_('Reset indexes'),
    view='indexing:index_instances_reset'
)

# Index template

link_index_template_create = Link(
    icon=icon_index_template_create,
    permissions=(permission_index_template_create,),
    text=_('Create index'), view='indexing:index_template_create'
)
link_index_template_delete = Link(
    args='resolved_object.pk', icon=icon_index_template_delete,
    permissions=(permission_index_template_delete,), tags='dangerous',
    text=_('Delete'), view='indexing:index_template_delete',
)
link_index_template_document_types = Link(
    args='resolved_object.pk', icon=icon_index_template_document_types,
    permissions=(permission_index_template_edit,),
    text=_('Document types'), view='indexing:index_template_document_types'
)
link_index_template_edit = Link(
    args='resolved_object.pk', icon=icon_index_template_edit,
    permissions=(permission_index_template_edit,), text=_('Edit'),
    view='indexing:index_template_edit'
)
link_index_template_event_triggers = Link(
    args='resolved_object.pk', icon=icon_index_template_event_triggers,
    permissions=(permission_index_template_edit,), text=_('Triggers'),
    view='indexing:index_template_event_triggers'
)
link_index_template_list = Link(
    icon=icon_index_template_list, text=_('Indexes'),
    view='indexing:index_template_list'
)
link_index_template_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='document_indexing', model_name='IndexTemplate',
        object_permission=permission_index_template_view,
        view_permission=permission_index_template_create,
    ), icon=icon_index, text=_('Indexes'),
    view='indexing:index_template_list'
)

# Index template node

link_index_template_node_tree_view = Link(
    args='resolved_object.pk', icon=icon_index_template_node_tree_view,
    permissions=(permission_index_template_edit,), text=_('Tree template'),
    view='indexing:index_template_view'
)
link_index_template_node_create = Link(
    args='resolved_object.pk', icon=icon_index_template_node_create,
    text=_('New child node'), view='indexing:template_node_create'
)
link_index_template_node_delete = Link(
    args='resolved_object.pk', condition=condition_is_not_root_node,
    icon=icon_index_template_node_delete, tags='dangerous',
    text=_('Delete'), view='indexing:template_node_delete'
)
link_index_template_node_edit = Link(
    args='resolved_object.pk', icon=icon_index_template_node_edit,
    condition=condition_is_not_root_node, text=_('Edit'),
    view='indexing:template_node_edit'
)



from django.contrib.contenttypes.models import ContentType

from mayan.apps.converter.layers import layer_saved_transformations
from mayan.apps.converter.permissions import (
    permission_transformation_delete, permission_transformation_edit
)
from mayan.apps.converter.tests.mixins import LayerTestMixin
from mayan.apps.documents.tests.literals import TEST_FILE_MULTI_PAGE_TIFF_FILENAME

from ..events import (
    event_document_version_page_created, event_document_version_page_deleted
)
from ..permissions import (
    permission_document_version_edit, permission_document_version_view
)

from .base import GenericDocumentViewTestCase
from .mixins.document_file_mixins import DocumentFileTestMixin
from .mixins.document_version_mixins import (
    DocumentVersionPageRemapViewTestMixin, DocumentVersionPageViewTestMixin,
    DocumentVersionTransformationTestMixin,
    DocumentVersionTransformationViewTestMixin
)


class DocumentVersionPageViewTestCase(
    DocumentVersionPageViewTestMixin, GenericDocumentViewTestCase
):
    def test_document_version_page_delete_no_permission(self):
        test_document_version_page_count = self._test_document_version.pages.count()

        self._clear_events()

        response = self._request_test_document_version_page_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document_version.pages.count(),
            test_document_version_page_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_delete_with_access(self):
        test_document_version_page_count = self._test_document_version.pages.count()

        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_edit
        )

        self._clear_events()

        response = self._request_test_document_version_page_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_document_version.pages.count(),
            test_document_version_page_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_version)
        self.assertEqual(
            events[0].verb, event_document_version_page_deleted.id
        )

    def test_trashed_document_version_page_delete_with_access(self):
        test_document_version_page_count = self._test_document_version.pages.count()

        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_edit
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document_version.pages.count(),
            test_document_version_page_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_list_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_version_page_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_list_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._clear_events()

        response = self._request_test_document_version_page_list_view()
        self.assertContains(
            response=response, status_code=200,
            text=str(self._test_document_version)
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_list_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_rotate_left_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_version_page_rotate_left_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_rotate_left_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._clear_events()

        response = self._request_test_document_version_page_rotate_left_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_rotate_left_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_rotate_left_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_rotate_right_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_version_page_rotate_right_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_rotate_right_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._clear_events()

        response = self._request_test_document_version_page_rotate_right_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_rotate_right_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_rotate_right_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_version_page_view(
            document_version_page=self._test_document_version.pages.first()
        )
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._clear_events()

        response = self._request_test_document_version_page_view(
            document_version_page=self._test_document_version.pages.first()
        )
        self.assertContains(
            response=response, status_code=200, text=str(
                self._test_document_version.pages.first()
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_view(
            document_version_page=self._test_document_version.pages.first()
        )
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_zoom_in_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_version_page_zoom_in_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_zoom_in_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._clear_events()

        response = self._request_test_document_version_page_zoom_in_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_zoom_in_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_zoom_in_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_zoom_out_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_version_page_zoom_out_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_zoom_out_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._clear_events()

        response = self._request_test_document_version_page_zoom_out_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_zoom_out_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_zoom_out_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentVersionPageRemapViewTestCase(
    DocumentVersionPageRemapViewTestMixin, DocumentFileTestMixin,
    GenericDocumentViewTestCase
):
    def setUp(self):
        super().setUp()
        self._upload_test_document_file()
        self._test_document_file_pages = []
        self.source_content_types = []
        self.source_object_ids = []

        for test_document_file in self._test_document.files.all():
            for test_document_file_page in test_document_file.pages.all():
                self._test_document_file_pages.append(test_document_file_page)
                self.source_content_types.append(
                    ContentType.objects.get_for_model(
                        model=test_document_file_page
                    )
                )
                self.source_object_ids.append(test_document_file_page.pk)

        self.single_page_remap_data = {
            'form-0-source_content_type': self.source_content_types[0].pk,
            'form-0-source_object_id': self.source_object_ids[0],
            'form-0-source_page_number': '1',
            'form-0-target_page_number': '1',
            'form-TOTAL_FORMS': '1',
            'form-INITIAL_FORMS': '0',
            'form-MAX_NUM_FORMS': '',
        }
        self.repeated_page_number_remap_data = {
            'form-0-source_content_type': self.source_content_types[0].pk,
            'form-0-source_object_id': self.source_object_ids[0],
            'form-0-source_page_number': '1',
            'form-0-target_page_number': '1',
            'form-1-source_content_type': self.source_content_types[1].pk,
            'form-1-source_object_id': self.source_object_ids[1],
            'form-1-source_page_number': '1',
            'form-1-target_page_number': '1',
            'form-TOTAL_FORMS': '2',
            'form-INITIAL_FORMS': '0',
            'form-MAX_NUM_FORMS': '',
        }

    def test_document_version_remap_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_version_page_list_remap_view(
            data=self.single_page_remap_data
        )
        self.assertEqual(response.status_code, 404)

        self._test_document_version.refresh_from_db()

        self.assertNotEqual(
            self._test_document_version.pages.first().content_object,
            self._test_document_file_pages[0]
        )
        self.assertEqual(
            self._test_document_version.pages.first().content_object,
            self._test_document_file_pages[1]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_remap_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_edit
        )

        self._clear_events()

        response = self._request_test_document_version_page_list_remap_view(
            data=self.single_page_remap_data
        )
        self.assertEqual(response.status_code, 302)

        self._test_document_version.refresh_from_db()

        self.assertEqual(
            self._test_document_version.pages.first().content_object,
            self._test_document_file_pages[0]
        )
        self.assertNotEqual(
            self._test_document_version.pages.first().content_object,
            self._test_document_file_pages[1]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_version)
        self.assertEqual(
            events[0].verb, event_document_version_page_deleted.id
        )

        self.assertEqual(events[1].action_object, self._test_document_version)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(
            events[1].target, self._test_document_version.pages[0]
        )
        self.assertEqual(
            events[1].verb, event_document_version_page_created.id
        )

    def test_trashed_document_version_remap_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_edit
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_page_list_remap_view(
            data=self.single_page_remap_data
        )
        self.assertEqual(response.status_code, 404)

        self._test_document_version.refresh_from_db()

        self.assertNotEqual(
            self._test_document_version.pages.first().content_object,
            self._test_document_file_pages[0]
        )
        self.assertEqual(
            self._test_document_version.pages.first().content_object,
            self._test_document_file_pages[1]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_remap_repeated_target_page_number_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_edit
        )

        self._clear_events()

        response = self._request_test_document_version_page_list_remap_view(
            data=self.repeated_page_number_remap_data
        )
        self.assertEqual(response.status_code, 200)

        self._test_document_version.refresh_from_db()

        self.assertNotEqual(
            self._test_document_version.pages.first().content_object,
            self._test_document_file_pages[0]
        )
        self.assertEqual(
            self._test_document_version.pages.first().content_object,
            self._test_document_file_pages[1]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentVersionTransformationViewTestCase(
    LayerTestMixin, DocumentVersionTransformationTestMixin,
    DocumentVersionTransformationViewTestMixin, GenericDocumentViewTestCase
):
    _test_document_filename = TEST_FILE_MULTI_PAGE_TIFF_FILENAME

    def setUp(self):
        super().setUp()
        self.assertTrue(
            expr=self._test_document_version.pages.count() > 1,
            msg='Test document must have more than one page'
        )

    def test_document_version_transformations_clear_view_no_permission(self):
        self._create_document_version_transformation()

        transformation_count = layer_saved_transformations.get_transformations_for(
            obj=self._test_document_version.pages.first()
        ).count()

        self._clear_events()

        response = self._request_test_document_version_transformations_clear_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            layer_saved_transformations.get_transformations_for(
                obj=self._test_document_version.pages.first()
            ).count(), transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_transformations_clear_view_with_access(self):
        self._create_document_version_transformation()

        transformation_count = layer_saved_transformations.get_transformations_for(
            obj=self._test_document_version.pages.first()
        ).count()

        self.grant_access(
            obj=self._test_document_version,
            permission=permission_transformation_delete
        )

        self._clear_events()

        response = self._request_test_document_version_transformations_clear_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            layer_saved_transformations.get_transformations_for(
                obj=self._test_document_version.pages.first()
            ).count(), transformation_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_transformations_clear_view_with_access(self):
        self._create_document_version_transformation()

        transformation_count = layer_saved_transformations.get_transformations_for(
            obj=self._test_document_version.pages.first()
        ).count()

        self.grant_access(
            obj=self._test_document_version,
            permission=permission_transformation_delete
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_transformations_clear_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            layer_saved_transformations.get_transformations_for(
                obj=self._test_document_version.pages.first()
            ).count(), transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_multiple_transformations_clear_view_no_permission(self):
        self._create_document_version_transformation()

        transformation_count = layer_saved_transformations.get_transformations_for(
            obj=self._test_document_version.pages.first()
        ).count()

        self._clear_events()

        response = self._request_test_document_version_multiple_transformations_clear_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            layer_saved_transformations.get_transformations_for(
                obj=self._test_document_version.pages.first()
            ).count(), transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_multiple_transformations_clear_view_with_access(self):
        self._create_document_version_transformation()

        transformation_count = layer_saved_transformations.get_transformations_for(
            obj=self._test_document_version.pages.first()
        ).count()

        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_view
        )
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_transformation_delete
        )

        self._clear_events()

        response = self._request_test_document_version_multiple_transformations_clear_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            layer_saved_transformations.get_transformations_for(
                obj=self._test_document_version.pages.first()
            ).count(), transformation_count - 1,
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_multiple_transformations_clear_view_with_access(self):
        self._create_document_version_transformation()

        transformation_count = layer_saved_transformations.get_transformations_for(
            obj=self._test_document_version.pages.first()
        ).count()

        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_view
        )
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_transformation_delete
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_multiple_transformations_clear_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            layer_saved_transformations.get_transformations_for(
                obj=self._test_document_version.pages.first()
            ).count(), transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_transformations_clone_view_no_permission(self):
        self._create_document_version_transformation()

        page_first_transformation_count = layer_saved_transformations.get_transformations_for(
            obj=self._test_document_version.pages.first()
        ).count()
        page_last_transformation_count = layer_saved_transformations.get_transformations_for(
            obj=self._test_document_version.pages.last()
        ).count()

        self._clear_events()

        response = self._request_test_document_version_transformations_clone_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            layer_saved_transformations.get_transformations_for(
                obj=self._test_document_version.pages.first()
            ).count(), page_first_transformation_count
        )
        self.assertEqual(
            layer_saved_transformations.get_transformations_for(
                obj=self._test_document_version.pages.last()
            ).count(), page_last_transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_transformations_clone_view_with_access(self):
        self._create_document_version_transformation()

        page_first_transformation_count = layer_saved_transformations.get_transformations_for(
            obj=self._test_document_version.pages.first()
        ).count()
        page_last_transformation_count = layer_saved_transformations.get_transformations_for(
            obj=self._test_document_version.pages.last()
        ).count()

        self.grant_access(
            obj=self._test_document_version,
            permission=permission_transformation_edit
        )

        self._clear_events()

        response = self._request_test_document_version_transformations_clone_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            layer_saved_transformations.get_transformations_for(
                obj=self._test_document_version.pages.first()
            ).count(), page_first_transformation_count
        )
        self.assertEqual(
            layer_saved_transformations.get_transformations_for(
                obj=self._test_document_version.pages.last()
            ).count(), page_last_transformation_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_transformations_clone_view_with_access(self):
        self._create_document_version_transformation()

        page_first_transformation_count = layer_saved_transformations.get_transformations_for(
            obj=self._test_document_version.pages.first()
        ).count()
        page_last_transformation_count = layer_saved_transformations.get_transformations_for(
            obj=self._test_document_version.pages.last()
        ).count()

        self.grant_access(
            obj=self._test_document_version,
            permission=permission_transformation_edit
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_version_transformations_clone_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            layer_saved_transformations.get_transformations_for(
                obj=self._test_document_version.pages.first()
            ).count(), page_first_transformation_count
        )
        self.assertEqual(
            layer_saved_transformations.get_transformations_for(
                obj=self._test_document_version.pages.last()
            ).count(), page_last_transformation_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.testing.tests.base import BaseTestCase

from .mixins import QuotaTestMixin


class QuotaModelTestCase(QuotaTestMixin, BaseTestCase):
    def test_method_get_absolute_url(self):
        self._create_test_quota()

        self._clear_events()

        self.assertTrue(
            self._test_quota.get_absolute_url()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from pathlib import Path

from django.conf import settings

DEFAULT_SOURCES_BACKEND_ARGUMENTS = {}

DEFAULT_SOURCES_CACHE_STORAGE_BACKEND = 'django.core.files.storage.FileSystemStorage'
DEFAULT_SOURCES_CACHE_STORAGE_BACKEND_ARGUMENTS = {
    'location': str(
        Path(settings.MEDIA_ROOT, 'source_cache')
    )
}

DEFAULT_SOURCES_LOCK_EXPIRE = 600

SOURCE_ACTION_EXECUTE_TASK_PATH = 'mayan.apps.sources.tasks.task_source_backend_action_execute'

STORAGE_NAME_SOURCE_CACHE_FOLDER = 'sources__source_cache'



from django.core import management

from ...utils import purge_periodic_tasks


class Command(management.BaseCommand):
    help = 'Removes all periodic tasks.'

    def handle(self, *args, **options):
        purge_periodic_tasks()



import requests

from django.conf import settings
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.authentication.classes import AuthenticationBackend
from mayan.apps.authentication.exceptions import AuthenticationError
from mayan.apps.common.classes import UpstreamSettingCollection
from mayan.apps.common.utils import get_class_full_name

from .django_authentication_backends import DjangoAuthenticationBackendOIDC
from .forms import AuthenticationFormOIDC
from .literals import (
    DEFAULT_OIDC_OP_AUTHORIZATION_ENDPOINT, DEFAULT_OIDC_OP_JWKS_ENDPOINT,
    DEFAULT_OIDC_OP_TOKEN_ENDPOINT, DEFAULT_OIDC_OP_USER_ENDPOINT,
    DEFAULT_OIDC_RP_CLIENT_ID, DEFAULT_OIDC_RP_CLIENT_SECRET,
    DEFAULT_OIDC_RP_SIGN_ALGO, DEFAULT_OIDC_USERNAME_ALGO
)


class AuthenticationBackendOIDC(AuthenticationBackend):
    login_form_class = AuthenticationFormOIDC

    def __init__(self, oidc_discovery_url=None, **kwargs):
        if oidc_discovery_url:
            response = requests.get(url=oidc_discovery_url)
            data = response.json()

            try:
                kwargs['oidc_op_authorization_endpoint'] = data['authorization_endpoint']
                kwargs['oidc_op_jwks_endpoint'] = data['jwks_uri']
                kwargs['oidc_op_token_endpoint'] = data['token_endpoint']
                kwargs['oidc_op_user_endpoint'] = data['userinfo_endpoint']
            except KeyError:
                raise AuthenticationError(
                    'Unable to extract OIDC provider discovery information.'
                )

        upstream_setting_collection = UpstreamSettingCollection(
            name='mozilla_django_oidc'
        )
        upstream_setting_collection.do_setting_add(
            default=DEFAULT_OIDC_OP_AUTHORIZATION_ENDPOINT,
            name='OIDC_OP_AUTHORIZATION_ENDPOINT'
        )
        upstream_setting_collection.do_setting_add(
            default=DEFAULT_OIDC_OP_JWKS_ENDPOINT,
            name='OIDC_OP_JWKS_ENDPOINT'
        )
        upstream_setting_collection.do_setting_add(
            default=DEFAULT_OIDC_OP_TOKEN_ENDPOINT,
            name='OIDC_OP_TOKEN_ENDPOINT'
        )
        upstream_setting_collection.do_setting_add(
            default=DEFAULT_OIDC_OP_USER_ENDPOINT,
            name='OIDC_OP_USER_ENDPOINT'
        )
        upstream_setting_collection.do_setting_add(
            default=DEFAULT_OIDC_RP_CLIENT_ID,
            name='OIDC_RP_CLIENT_ID'
        )
        upstream_setting_collection.do_setting_add(
            default=DEFAULT_OIDC_RP_CLIENT_SECRET,
            name='OIDC_RP_CLIENT_SECRET'
        )
        upstream_setting_collection.do_setting_add(
            default=DEFAULT_OIDC_RP_SIGN_ALGO,
            name='OIDC_RP_SIGN_ALGO'
        )
        upstream_setting_collection.do_setting_add(
            default=DEFAULT_OIDC_USERNAME_ALGO,
            name='OIDC_USERNAME_ALGO'
        )

        upstream_setting_collection.do_kwargs_capture(**kwargs)

        super().__init__()

    def initialize(self):
        settings.AUTHENTICATION_BACKENDS = (
            get_class_full_name(klass=DjangoAuthenticationBackendOIDC),
        )

        settings.MIDDLEWARE += (
            'mozilla_django_oidc.middleware.SessionRefresh',
        )

        settings.REST_FRAMEWORK['DEFAULT_AUTHENTICATION_CLASSES'] = (
            'mozilla_django_oidc.contrib.drf.OIDCAuthentication',
        ) + settings.REST_FRAMEWORK['DEFAULT_AUTHENTICATION_CLASSES']

    def get_context_data(self):
        return {
            'form_action': reverse(viewname='oidc_authentication_init'),
            'submit_button_css_extra_classes': 'center-block',
            'submit_label': _('Login via OIDC'),
            'submit_method': 'get'
        }



from rest_framework.generics import get_object_or_404

from mayan.apps.acls.models import AccessControlList
from mayan.apps.rest_api import generics

from ..models.document_models import Document
from ..models.favorite_document_models import FavoriteDocument
from ..permissions import permission_document_view
from ..serializers.favorite_document_serializers import (
    FavoriteDocumentSerializer
)


class APIFavoriteDocumentDetailView(generics.RetrieveDestroyAPIView):
    """
    delete: Delete the selected favorite document.
    get: Return the details of the selected favorite document.
    """
    lookup_url_kwarg = 'favorite_document_id'
    mayan_object_permissions = {
        'DELETE': (permission_document_view,),
        'GET': (permission_document_view,)
    }
    serializer_class = FavoriteDocumentSerializer

    def get_source_queryset(self):
        return FavoriteDocument.valid.filter(user=self.request.user)


class APIFavoriteDocumentListView(generics.ListCreateAPIView):
    """
    get: Return a list of the favorite documents for the current user.
    post: Add a new document to the list of favorite documents for the current user.
    """
    mayan_object_permissions = {
        'GET': (permission_document_view,)
    }
    serializer_class = FavoriteDocumentSerializer

    def get_instance_extra_data(self):
        return {
            'user': self.request.user
        }

    def get_source_queryset(self):
        return FavoriteDocument.valid.filter(user=self.request.user)

    def perform_create(self, serializer):
        queryset = Document.valid.all()

        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_document_view, queryset=queryset,
            user=self.request.user
        )

        serializer.validated_data['document'] = get_object_or_404(
            queryset=queryset, pk=serializer.validated_data['document_id']
        )
        super().perform_create(serializer=serializer)



import io
import logging
import shutil
import struct

from PIL import Image
import pypdf
import sh

from django.utils.translation import ugettext_lazy as _

from mayan.apps.storage.utils import NamedTemporaryFile

from ..classes import ConverterBase
from ..exceptions import PageCountError
from ..literals import (
    DEFAULT_PDFTOPPM_DPI, DEFAULT_PDFTOPPM_FORMAT, DEFAULT_PDFTOPPM_PATH,
    DEFAULT_PDFINFO_PATH, DEFAULT_PILLOW_MAXIMUM_IMAGE_PIXELS
)
from ..settings import setting_graphics_backend_arguments

logger = logging.getLogger(name=__name__)
pdftoppm_path = setting_graphics_backend_arguments.value.get(
    'pdftoppm_path', DEFAULT_PDFTOPPM_PATH
)

try:
    command_pdftoppm = sh.Command(path=pdftoppm_path)
except sh.CommandNotFound:
    command_pdftoppm = None
else:
    pdftoppm_format = '-{}'.format(
        setting_graphics_backend_arguments.value.get(
            'pdftoppm_format', DEFAULT_PDFTOPPM_FORMAT
        )
    )

    pdftoppm_dpi = format(
        setting_graphics_backend_arguments.value.get(
            'pdftoppm_dpi', DEFAULT_PDFTOPPM_DPI
        )
    )

    command_pdftoppm = command_pdftoppm.bake(
        pdftoppm_format, '-r', pdftoppm_dpi
    )

pdfinfo_path = setting_graphics_backend_arguments.value.get(
    'pdfinfo_path', DEFAULT_PDFINFO_PATH
)

try:
    command_pdfinfo = sh.Command(path=pdfinfo_path)
except sh.CommandNotFound:
    command_pdfinfo = None


pillow_maximum_image_pixels = setting_graphics_backend_arguments.value.get(
    'pillow_maximum_image_pixels', DEFAULT_PILLOW_MAXIMUM_IMAGE_PIXELS
)
Image.MAX_IMAGE_PIXELS = pillow_maximum_image_pixels


class Python(ConverterBase):
    def convert(self, *args, **kwargs):
        super().convert(*args, **kwargs)

        if self.mime_type == 'application/pdf' and command_pdftoppm:
            with NamedTemporaryFile() as new_file_object:
                self.file_object.seek(0)
                shutil.copyfileobj(
                    fsrc=self.file_object, fdst=new_file_object
                )
                self.file_object.seek(0)
                new_file_object.seek(0)

                image_buffer = io.BytesIO()
                command_pdftoppm(
                    new_file_object.name, f=self.page_number + 1,
                    l=self.page_number + 1, _out=image_buffer
                )
                image_buffer.seek(0)
                return Image.open(fp=image_buffer)

    def get_page_count(self):
        super().get_page_count()

        if self.mime_type == 'application/pdf' or self.soffice_file:
            if self.soffice_file:
                file_object = self.soffice_file
            else:
                file_object = self.file_object

            return self.get_pypdf_page_count(file_object=file_object)
        else:
            return self.get_pillow_page_count()

    def get_pillow_page_count(self):
        page_count = 1

        try:
            image = Image.open(fp=self.file_object)
        except IOError as exception:
            error_message = _(
                'Exception determining page count using Pillow; %s'
            ) % exception
            logger.error(error_message)
            raise PageCountError(error_message)
        finally:
            self.file_object.seek(0)

        # Get total page count by attempting to seek to an increasing
        # page count number until an EOFError or struct.error exception
        # are raised.
        while True:
            try:
                image.seek(
                    image.tell() + 1
                )
            except EOFError:
                """End of sequence"""
                break
            except struct.error:
                """
                struct.error was raise for a TIFF file converted to JPEG
                GitLab issue #767 "Upload Error: unpack_from requires a
                buffer of at least 2 bytes"
                """
                logger.debug(
                    msg='image page count detection raised struct.error'
                )
                break
            else:
                try:
                    # Even if the image reports multiple frames,
                    # test it to make sure it is valid and supported
                    # before counting it as a valid page.
                    image.getim()
                except OSError as exception:
                    logger.error(
                        'Multi image element not supported; %s',
                        exception
                    )
                    break
                else:
                    page_count += 1

        self.file_object.seek(0)
        return page_count

    def get_pdfinfo_page_count(self, file_object):
        process = command_pdfinfo('-', _in=file_object)

        for line in str(process.stdout).split('\n'):
            if line.startswith('Pages:'):
                line = line.replace('Pages:', '')
                page_count = int(line)
                break

        file_object.seek(0)
        logger.debug('Document contains %d pages', page_count)
        return page_count

    def get_pypdf_page_count(self, file_object):
        try:
            # Try PyPDF to determine the page count.
            pdf_reader = pypdf.PdfReader(
                stream=file_object, strict=False
            )
            page_count = len(pdf_reader.pages)
        except Exception as exception:
            if str(exception) == 'File has not been decrypted':
                # File is encrypted, try to decrypt using a blank
                # password.
                file_object.seek(0)
                pdf_reader = pypdf.PdfReader(
                    stream=file_object, strict=False
                )
                try:
                    pdf_reader.decrypt(password=b'')
                    page_count = len(pdf_reader.pages)
                except Exception as exception:
                    file_object.seek(0)
                    if str(exception) == 'only algorithm code 1 and 2 are supported':
                        # PDF uses an unsupported encryption.
                        # Try poppler-util's pdfinfo.
                        page_count = self.get_pdfinfo_page_count(
                            file_object=file_object
                        )
                        return page_count
                    else:
                        error_message = _(
                            'Exception determining PDF page count; %s'
                        ) % exception
                        logger.error(error_message, exc_info=True)
                        raise PageCountError(error_message)
            elif str(exception) == 'EOF marker not found':
                # PyPDF2 issue: https://github.com/mstamy2/PyPDF2/issues/177
                # Try poppler-util's pdfinfo.
                logger.debug(
                    msg='PyPDF2 GitHub issue #177 : EOF marker not found'
                )
                file_object.seek(0)
                page_count = self.get_pdfinfo_page_count(
                    file_object=file_object
                )
                return page_count
            else:
                error_message = _(
                    'Exception determining PDF page count; %s'
                ) % exception
                logger.error(error_message, exc_info=True)
                raise PageCountError(error_message)
        else:
            logger.debug('Document contains %d pages', page_count)
            return page_count
        finally:
            file_object.seek(0)



from django.apps import apps
from django.utils.translation import ugettext_lazy as _

from celery.schedules import crontab

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.celery import app

from .renderers import (
    RendererChartJSDoughnut, RendererChartJSLine, RendererChartJSPie
)


class StatisticNamespace(AppsModuleLoaderMixin):
    _loader_module_name = 'statistics'
    _registry = {}

    @classmethod
    def get_all(cls):
        return list(
            cls._registry.values()
        )

    @classmethod
    def get(cls, slug):
        return cls._registry[slug]

    def __init__(self, slug, label):
        self.slug = slug
        self.label = label
        self._statistics = []
        self.__class__._registry[slug] = self

    def __str__(self):
        return str(self.label)

    def add_statistic(self, klass, *args, **kwargs):
        statistic = klass(*args, **kwargs)
        statistic.namespace = self
        self._statistics.append(statistic)

    @property
    def statistics(self):
        return self._statistics


StatisticNamespace.verbose_name = _('Statistics namespace')


class StatisticType:
    _registry = {}
    renderer = None

    @staticmethod
    def evaluate(data):
        try:
            for key, value in data.items():
                return {
                    key: StatisticType.evaluate(data=value)
                }
        except AttributeError:
            if type(data) is map:
                data = list(data)

        return data

    @staticmethod
    def purge_schedules():
        PeriodicTask = apps.get_model(
            app_label='django_celery_beat', model_name='PeriodicTask'
        )
        StatisticResult = apps.get_model(
            app_label='mayan_statistics', model_name='StatisticResult'
        )

        queryset = PeriodicTask.objects.filter(
            name__startswith='mayan_statistics.'
        ).exclude(
            name__in=StatisticType.get_task_names()
        )

        for periodic_task in queryset:
            crontab_instance = periodic_task.crontab
            periodic_task.delete()

            if crontab_instance and not crontab_instance.periodictask_set.all():
                # Only delete the interval if nobody else is using it
                crontab_instance.delete()

        StatisticResult.objects.filter(
            slug__in=queryset.values_list('name', flat=True)
        ).delete()

    @classmethod
    def get(cls, slug):
        return cls._registry[slug]

    @classmethod
    def get_all(cls):
        return list(
            cls._registry.values()
        )

    @classmethod
    def get_task_names(cls):
        return [
            task.get_task_name() for task in cls.get_all()
        ]

    def __init__(
        self, func, label, slug, day_of_month='*', day_of_week='*',
        hour='*', minute='*', month_of_year='*'
    ):
        # Hidden import.
        from .queues import queue_statistics, task_execute_statistic

        self.slug = slug
        self.label = label
        self.func = func

        self.schedule = crontab(
            day_of_month=day_of_month, day_of_week=day_of_week, hour=hour,
            minute=minute, month_of_year=month_of_year
        )

        app.conf.beat_schedule.update(
            {
                self.get_task_name(): {
                    'task': task_execute_statistic.dotted_path,
                    'schedule': self.schedule,
                    'args': (self.slug,)
                }
            }
        )

        app.conf.task_routes.update(
            {
                self.get_task_name(): {
                    'queue': queue_statistics.name
                }
            }
        )

        self.__class__._registry[slug] = self

    def __str__(self):
        return str(self.label)

    def execute(self):
        results = self.func()
        # Force evaluation of results to be able to store it serialized.
        results = StatisticType.evaluate(data=results)
        self.store_results(results=results)

    def get_chart_context(self):
        return self.renderer(
            data=self.get_results_data()
        ).get_chart_context()

    def get_last_update(self):
        results = self.get_results()

        if results:
            return results.datetime
        else:
            return _('Never')

    def get_results(self, only=None):
        StatisticResult = apps.get_model(
            app_label='mayan_statistics', model_name='StatisticResult'
        )

        try:
            return StatisticResult.objects.get(slug=self.slug)
        except StatisticResult.DoesNotExist:
            return StatisticResult.objects.none()
        except StatisticResult.MultipleObjectsReturned:
            # This should not happen. Self-heal by deleting the duplicate
            # results.
            StatisticResult.objects.filter(slug=self.slug).delete()
            return StatisticResult.objects.none()

    def get_results_data(self):
        results = self.get_results()

        if results:
            return results.get_data()
        else:
            return {
                'series': {}
            }

    def get_task_name(self):
        return 'mayan_statistics.task_execute_statistic_{}'.format(self.slug)

    def store_results(self, results):
        StatisticResult = apps.get_model(
            app_label='mayan_statistics', model_name='StatisticResult'
        )

        StatisticResult.objects.filter(slug=self.slug).delete()

        statistic_result, created = StatisticResult.objects.get_or_create(
            slug=self.slug
        )
        statistic_result.store_data(data=results)


class StatisticTypeDoughnutChart(StatisticType):
    renderer = RendererChartJSDoughnut
    type_label = _('Doughnut chart')


class StatisticTypeLineChart(StatisticType):
    renderer = RendererChartJSLine
    type_label = _('Line chart')


class StatisticTypePieChart(StatisticType):
    renderer = RendererChartJSPie
    type_label = _('Pie chart')



from django.test import override_settings

from mayan.apps.smart_settings.tests.mixins import SmartSettingTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..literals import (
    DEFAULT_DOCUMENTS_STORAGE_BACKEND,
    DEFAULT_DOCUMENTS_STORAGE_BACKEND_ARGUMENTS
)
from ..settings import (
    setting_document_file_page_image_cache_storage_backend,
    setting_document_file_page_image_cache_storage_backend_arguments,
    setting_document_file_storage_backend,
    setting_document_file_storage_backend_arguments,
    setting_document_version_page_image_cache_storage_backend,
    setting_document_version_page_image_cache_storage_backend_arguments,
    setting_recently_accessed_document_count,
    setting_recently_created_document_count
)

from .literals import (
    TEST_DOCUMENTS_CACHE_STORAGE_BACKEND,
    TEST_DOCUMENTS_CACHE_STORAGE_BACKEND_ARGUMENTS,
    TEST_DOCUMENTS_RECENTLY_ACCESSED_COUNT,
    TEST_DOCUMENTS_RECENTLY_CREATED_COUNT, TEST_DOCUMENTS_STORAGE_BACKEND,
    TEST_DOCUMENTS_STORAGE_BACKEND_ARGUMENTS
)


class DocumentSettingMigrationTestCase(SmartSettingTestMixin, BaseTestCase):
    @override_settings(DOCUMENTS_CACHE_STORAGE_BACKEND=TEST_DOCUMENTS_CACHE_STORAGE_BACKEND)
    def test_setting_document_file_page_image_cache_storage_backend_0003(self):
        test_value = None
        self.test_setting = setting_document_file_page_image_cache_storage_backend
        self.test_config_value = test_value
        self._create_test_config_file()

        self.assertEqual(
            setting_document_file_page_image_cache_storage_backend.value,
            TEST_DOCUMENTS_CACHE_STORAGE_BACKEND
        )

    @override_settings(DOCUMENTS_CACHE_STORAGE_BACKEND_ARGUMENTS=TEST_DOCUMENTS_CACHE_STORAGE_BACKEND_ARGUMENTS)
    def test_setting_document_file_page_image_cache_storage_backend_arguments_0003(self):
        test_value = None
        self.test_setting = setting_document_file_page_image_cache_storage_backend_arguments
        self.test_config_value = test_value
        self._create_test_config_file()

        self.assertEqual(
            setting_document_file_page_image_cache_storage_backend_arguments.value,
            TEST_DOCUMENTS_CACHE_STORAGE_BACKEND_ARGUMENTS
        )

    @override_settings(DOCUMENTS_STORAGE_BACKEND=TEST_DOCUMENTS_STORAGE_BACKEND)
    def test_setting_document_file_storage_backend_0003(self):
        test_value = None
        self.test_setting = setting_document_file_storage_backend
        self.test_config_value = test_value
        self._create_test_config_file()

        self.assertEqual(
            setting_document_file_storage_backend.value,
            TEST_DOCUMENTS_STORAGE_BACKEND
        )

    def test_setting_document_file_storage_backend_0003_no_value(self):
        test_value = None
        self.test_setting = setting_document_file_storage_backend
        self.test_config_value = test_value
        self._create_test_config_file()

        self.assertEqual(
            setting_document_file_storage_backend.value,
            DEFAULT_DOCUMENTS_STORAGE_BACKEND
        )

    @override_settings(DOCUMENTS_STORAGE_BACKEND_ARGUMENTS=TEST_DOCUMENTS_STORAGE_BACKEND_ARGUMENTS)
    def test_setting_document_file_storage_backend_arguments_0003(self):
        test_value = None
        self.test_setting = setting_document_file_storage_backend_arguments
        self.test_config_value = test_value
        self._create_test_config_file()

        self.assertEqual(
            setting_document_file_storage_backend_arguments.value,
            TEST_DOCUMENTS_STORAGE_BACKEND_ARGUMENTS
        )

    def test_setting_document_file_storage_backend_arguments_0003_no_value(self):
        test_value = None
        self.test_setting = setting_document_file_storage_backend_arguments
        self.test_config_value = test_value
        self._create_test_config_file()

        self.assertEqual(
            setting_document_file_storage_backend_arguments.value,
            DEFAULT_DOCUMENTS_STORAGE_BACKEND_ARGUMENTS
        )

    @override_settings(DOCUMENTS_CACHE_STORAGE_BACKEND=TEST_DOCUMENTS_CACHE_STORAGE_BACKEND)
    def test_setting_document_version_page_image_cache_storage_backend_0003(self):
        test_value = None
        self.test_setting = setting_document_version_page_image_cache_storage_backend
        self.test_config_value = test_value
        self._create_test_config_file()

        self.assertEqual(
            setting_document_version_page_image_cache_storage_backend.value,
            TEST_DOCUMENTS_CACHE_STORAGE_BACKEND
        )

    @override_settings(DOCUMENTS_CACHE_STORAGE_BACKEND_ARGUMENTS=TEST_DOCUMENTS_CACHE_STORAGE_BACKEND_ARGUMENTS)
    def test_setting_document_version_page_image_cache_storage_backend_arguments_0003(self):
        test_value = None
        self.test_setting = setting_document_version_page_image_cache_storage_backend_arguments
        self.test_config_value = test_value
        self._create_test_config_file()

        self.assertEqual(
            setting_document_version_page_image_cache_storage_backend_arguments.value,
            TEST_DOCUMENTS_CACHE_STORAGE_BACKEND_ARGUMENTS
        )

    @override_settings(DOCUMENTS_RECENT_ACCESS_COUNT=TEST_DOCUMENTS_RECENTLY_ACCESSED_COUNT)
    def test_setting_recently_accessed_document_count_0002(self):
        test_value = None
        self.test_setting = setting_recently_accessed_document_count
        self.test_config_value = test_value
        self._create_test_config_file()

        self.assertEqual(
            setting_recently_accessed_document_count.value,
            TEST_DOCUMENTS_RECENTLY_ACCESSED_COUNT
        )

    @override_settings(DOCUMENTS_RECENT_ADDED_COUNT=TEST_DOCUMENTS_RECENTLY_CREATED_COUNT)
    def test_setting_recently_created_document_count_0002(self):
        test_value = None
        self.test_setting = setting_recently_created_document_count
        self.test_config_value = test_value
        self._create_test_config_file()

        self.assertEqual(
            setting_recently_created_document_count.value,
            TEST_DOCUMENTS_RECENTLY_CREATED_COUNT
        )



from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL)
    ]

    operations = [
        migrations.CreateModel(
            name='RecentSearch',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'query', models.TextField(
                        verbose_name='Query', editable=False
                    )
                ),
                (
                    'datetime_created', models.DateTimeField(
                        auto_now=True, verbose_name='Datetime created',
                        db_index=True
                    )
                ),
                (
                    'hits', models.IntegerField(
                        verbose_name='Hits', editable=False
                    )
                ),
                (
                    'user', models.ForeignKey(
                        on_delete=models.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name='User'
                    )
                )
            ],
            options={
                'ordering': ('-datetime_created',),
                'verbose_name': 'Recent search',
                'verbose_name_plural': 'Recent searches'
            },
            bases=(models.Model,)
        )
    ]



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_credential_created, event_credential_edited
from ..models import StoredCredential
from ..permissions import (
    permission_credential_create, permission_credential_delete,
    permission_credential_edit, permission_credential_view
)
from .mixins import StoredCredentialAPIViewTestMixin


class StoredCredentialAPIViewTestCase(
    StoredCredentialAPIViewTestMixin, BaseAPITestCase
):
    auto_create_test_credential = False

    def test_credential_create_api_view_no_permission(self):
        credential_count = StoredCredential.objects.count()

        self._clear_events()

        response = self._request_test_stored_credential_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(StoredCredential.objects.count(), credential_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_credential_create_api_view_with_access(self):
        credential_count = StoredCredential.objects.count()

        self.grant_permission(permission=permission_credential_create)

        self._clear_events()

        response = self._request_test_stored_credential_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(
            StoredCredential.objects.count(), credential_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_stored_credential)
        self.assertEqual(events[0].verb, event_credential_created.id)

    def test_credential_delete_api_view_no_permission(self):
        self._create_test_stored_credential()

        credential_count = StoredCredential.objects.count()

        self._clear_events()

        response = self._request_test_stored_credential_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            StoredCredential.objects.count(), credential_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_credential_delete_api_view_with_access(self):
        self._create_test_stored_credential()

        credential_count = StoredCredential.objects.count()

        self.grant_access(
            obj=self._test_stored_credential,
            permission=permission_credential_delete
        )

        self._clear_events()

        response = self._request_test_stored_credential_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(
            StoredCredential.objects.count(), credential_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_credential_detail_api_view_no_permission(self):
        self._create_test_stored_credential()

        self._clear_events()

        response = self._request_test_stored_credential_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('id' not in response.data)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_credential_detail_api_view_with_access(self):
        self._create_test_stored_credential()

        self.grant_access(
            obj=self._test_stored_credential,
            permission=permission_credential_view
        )

        self._clear_events()

        response = self._request_test_stored_credential_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['id'], self._test_stored_credential.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_credential_edit_via_patch_api_view_no_permssion(self):
        self._create_test_stored_credential()

        credential_label = self._test_stored_credential.label

        self._clear_events()

        response = self._request_test_stored_credential_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_stored_credential.refresh_from_db()
        self.assertEqual(
            self._test_stored_credential.label, credential_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_credential_edit_via_patch_api_view_with_access(self):
        self._create_test_stored_credential()

        credential_label = self._test_stored_credential.label

        self.grant_access(
            obj=self._test_stored_credential,
            permission=permission_credential_edit
        )

        self._clear_events()

        response = self._request_test_stored_credential_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_stored_credential.refresh_from_db()
        self.assertNotEqual(
            self._test_stored_credential.label, credential_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_stored_credential)
        self.assertEqual(events[0].verb, event_credential_edited.id)

    def test_credential_list_api_view_no_permission(self):
        self._create_test_stored_credential()

        self._clear_events()

        response = self._request_test_stored_credential_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_credential_list_api_view_with_access(self):
        self._create_test_stored_credential()

        self.grant_access(
            obj=self._test_stored_credential,
            permission=permission_credential_view
        )

        self._clear_events()

        response = self._request_test_stored_credential_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['id'],
            self._test_stored_credential.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link

from ..icons import (
    icon_document_trash_send, icon_trash_can_empty,
    icon_trashed_document_delete, icon_trashed_document_list,
    icon_trashed_document_restore
)
from ..permissions import (
    permission_document_trash, permission_trash_empty,
    permission_trashed_document_delete, permission_trashed_document_restore
)

link_document_delete = Link(
    args='resolved_object.id', icon=icon_trashed_document_delete,
    permissions=(permission_trashed_document_delete,),
    tags='dangerous', text=_('Delete'), view='documents:document_delete'
)
link_document_trash = Link(
    args='resolved_object.id', icon=icon_document_trash_send,
    permissions=(permission_document_trash,), tags='dangerous',
    text=_('Move to trash'), view='documents:document_trash'
)
link_document_list_deleted = Link(
    icon=icon_trashed_document_list, text=_('Trash can'),
    view='documents:document_list_deleted'
)
link_document_restore = Link(
    args='object.pk', icon=icon_trashed_document_restore,
    permissions=(permission_trashed_document_restore,), text=_('Restore'),
    view='documents:document_restore'
)
link_document_multiple_trash = Link(
    icon=icon_document_trash_send, tags='dangerous',
    text=_('Move to trash'), view='documents:document_multiple_trash'
)
link_document_multiple_delete = Link(
    icon=icon_trashed_document_delete, tags='dangerous',
    text=_('Delete'), view='documents:document_multiple_delete'
)
link_document_multiple_restore = Link(
    icon=icon_trashed_document_restore, text=_('Restore'),
    view='documents:document_multiple_restore'
)
link_trash_can_empty = Link(
    icon=icon_trash_can_empty, permissions=(permission_trash_empty,),
    text=_('Empty trash'), view='documents:trash_can_empty'
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_('Authentication OTP'), name='authentication_otp'
)

event_otp_disabled = namespace.add_event_type(
    label=_('OTP disabled'), name='otp_disabled'
)
event_otp_enabled = namespace.add_event_type(
    label=_('OTP enabled'), name='otp_enabled'
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.serializers import ContentTypeSerializer
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import MultiKwargHyperlinkedIdentityField

from .models import ErrorLogPartitionEntry


class ErrorLogPartitionEntrySerializer(serializers.ModelSerializer):
    content_type = ContentTypeSerializer(
        label=_('Content type'), read_only=True,
        source='error_log_partition.content_type'
    )
    object_id = serializers.IntegerField(
        label=_('Object ID'), source='error_log_partition.object_id'
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_('URL'), view_kwargs=(
            {
                'lookup_field': 'error_log_partition.content_type.app_label',
                'lookup_url_kwarg': 'app_label'
            },
            {
                'lookup_field': 'error_log_partition.content_type.model',
                'lookup_url_kwarg': 'model_name'
            },
            {
                'lookup_field': 'error_log_partition.object_id',
                'lookup_url_kwarg': 'object_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'error_log_partition_entry_id'
            }
        ), view_name='rest_api:errorlogpartitionentry-detail'
    )

    class Meta:
        fields = (
            'content_type', 'datetime', 'id', 'object_id', 'text', 'url'
        )
        model = ErrorLogPartitionEntry
        read_only_fields = fields



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import (
    icon_queue_task_type_list, icon_worker_list, icon_worker_queue_list
)
from .permissions import permission_task_view

link_queue_task_type_list = Link(
    icon=icon_queue_task_type_list, kwargs={
        'queue_name': 'resolved_object.name'
    }, permissions=(permission_task_view,), text=_('Task type list'),
    view='task_manager:queue_task_type_list'
)
link_worker_list = Link(
    icon=icon_worker_list, permissions=(permission_task_view,),
    text=_('Worker list'), view='task_manager:worker_list'
)
link_worker_queue_list = Link(
    icon=icon_worker_queue_list, kwargs={
        'worker_name': 'resolved_object.name'
    }, permissions=(permission_task_view,), text=_('Queue list'),
    view='task_manager:worker_queue_list'
)



import logging

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.urls import reverse
from django.utils.translation import ugettext, ugettext_lazy as _

from mayan.apps.common.validators import (
    YAMLValidator, validate_internal_name
)
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave
from mayan.apps.storage.classes import DefinedStorageLazy

from .classes import Layer
from .events import event_asset_created, event_asset_edited
from .literals import STORAGE_NAME_ASSETS
from .managers import LayerTransformationManager, ObjectLayerManager
from .model_mixins import (
    AssetBusinessLogicMixin, ObjectLayerBusinessLogicMixin,
    LayerTransformationBusinessLogicMixin
)
from .transformations import BaseTransformation
from .utils import model_upload_to

logger = logging.getLogger(name=__name__)


class Asset(AssetBusinessLogicMixin, ExtraDataModelMixin, models.Model):
    """
    This model keeps track of files that will be available for use with
    transformations.
    """
    label = models.CharField(
        max_length=96, unique=True, verbose_name=_('Label')
    )
    internal_name = models.CharField(
        db_index=True, help_text=_(
            'This value will be used when referencing this asset. '
            'Can only contain letters, numbers, and underscores.'
        ), max_length=255, unique=True, validators=[validate_internal_name],
        verbose_name=_('Internal name')
    )
    file = models.FileField(
        storage=DefinedStorageLazy(name=STORAGE_NAME_ASSETS),
        upload_to=model_upload_to, verbose_name=_('File')
    )

    class Meta:
        ordering = ('label',)
        verbose_name = _('Asset')
        verbose_name_plural = _('Assets')

    def __str__(self):
        return self.label

    def delete(self, *args, **kwargs):
        self.cache_partition.delete()
        name = self.file.name
        self.file.close()
        self.file.storage.delete(name=name)
        return super().delete(*args, **kwargs)

    def get_absolute_url(self):
        return reverse(
            viewname='converter:asset_detail', kwargs={
                'asset_id': self.pk
            }
        )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_asset_created,
            'target': 'self'
        },
        edited={
            'event': event_asset_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class StoredLayer(models.Model):
    name = models.CharField(
        max_length=64, unique=True, verbose_name=_('Name')
    )
    order = models.PositiveIntegerField(
        db_index=True, unique=True, verbose_name=_('Order')
    )

    class Meta:
        ordering = ('order',)
        verbose_name = _('Stored layer')
        verbose_name_plural = _('Stored layers')

    def __str__(self):
        return self.name

    def get_layer(self):
        return Layer.get(name=self.name)


class ObjectLayer(ObjectLayerBusinessLogicMixin, models.Model):
    content_type = models.ForeignKey(
        on_delete=models.CASCADE, to=ContentType,
        verbose_name=_('Content type')
    )
    object_id = models.PositiveIntegerField(
        verbose_name=_('Object ID')
    )
    content_object = GenericForeignKey(
        ct_field='content_type', fk_field='object_id'
    )
    enabled = models.BooleanField(
        default=True, verbose_name=_('Enabled')
    )
    stored_layer = models.ForeignKey(
        on_delete=models.CASCADE, related_name='object_layers',
        to=StoredLayer, verbose_name=_('Stored layer')
    )

    objects = ObjectLayerManager()

    class Meta:
        ordering = ('stored_layer__order',)
        unique_together = ('content_type', 'object_id', 'stored_layer')
        verbose_name = _('Object layer')
        verbose_name_plural = _('Object layers')


class LayerTransformation(
    LayerTransformationBusinessLogicMixin, models.Model
):
    """
    Model that stores the transformation and transformation arguments
    for a given object
    Fields:
    * order - Order of a Transformation - In case there are multiple
    transformations for an object, this field list the order at which
    they will be execute.
    * arguments - Arguments of a Transformation - An optional field to hold a
    transformation argument. Example: if a page is rotated with the Rotation
    transformation, this field will show by how many degrees it was rotated.
    """
    object_layer = models.ForeignKey(
        on_delete=models.CASCADE, related_name='transformations',
        to=ObjectLayer, verbose_name=_('Object layer')
    )
    order = models.PositiveIntegerField(
        blank=True, db_index=True, default=0, help_text=_(
            'Order in which the transformations will be executed. If left '
            'unchanged, an automatic order value will be assigned.'
        ), verbose_name=_('Order')
    )
    name = models.CharField(
        max_length=128, verbose_name=_('Name')
    )
    arguments = models.TextField(
        blank=True, help_text=_(
            'Enter the arguments for the transformation as a YAML '
            'dictionary. ie: {"degrees": 180}'
        ), validators=[
            YAMLValidator()
        ], verbose_name=_('Arguments')
    )
    enabled = models.BooleanField(
        default=True, verbose_name=_('Enabled')
    )

    objects = LayerTransformationManager()

    class Meta:
        ordering = ('object_layer__stored_layer__order', 'order',)
        unique_together = ('object_layer', 'order')
        verbose_name = _('Layer transformation')
        verbose_name_plural = _('Layer transformations')

    def __str__(self):
        try:
            return str(
                BaseTransformation.get(name=self.name)
            )
        except KeyError:
            return ugettext('Unknown transformation class')

    def save(self, *args, **kwargs):
        if not self.order:
            self.order = self.object_layer.get_next_order()
        super().save(*args, **kwargs)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import (
    icon_statistic_detail, icon_statistic_namespace_detail,
    icon_statistic_namespace_list, icon_statistic_queue, icon_statistics
)
from .permissions import permission_statistics_view

# Translators: 'Queue' here is the verb, to queue a statistic to update.
link_statistic_namespace_detail = Link(
    args='resolved_object.slug', icon=icon_statistic_namespace_detail,
    permissions=(permission_statistics_view,), text=_('Namespace details'),
    view='statistics:statistic_namespace_detail'
)
link_statistic_namespace_list = Link(
    icon=icon_statistic_namespace_list,
    permissions=(permission_statistics_view,), text=_('Namespace list'),
    view='statistics:statistic_namespace_list'
)
link_statistic_type_queue = Link(
    args='resolved_object.slug', icon=icon_statistic_queue,
    permissions=(permission_statistics_view,), text=_('Queue'),
    view='statistics:statistic_queue'
)
link_statistic_type_detail = Link(
    args='resolved_object.slug', icon=icon_statistic_detail,
    permissions=(permission_statistics_view,), text=_('View'),
    view='statistics:statistic_detail'
)
link_statistics = Link(
    icon=icon_statistics, permissions=(permission_statistics_view,),
    text=_('Statistics'), view='statistics:statistic_namespace_list'
)



from django.utils.translation import ugettext_lazy as _

from rest_framework.exceptions import ValidationError
from rest_framework.reverse import reverse

from mayan.apps.events.classes import EventType
from mayan.apps.events.serializers import EventTypeSerializer
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import (
    FilteredPrimaryKeyRelatedField, MultiKwargHyperlinkedIdentityField
)

from ..models.workflow_transition_models import (
    WorkflowTransition, WorkflowTransitionField,
    WorkflowTransitionTriggerEvent
)

from .workflow_template_state_serializers import WorkflowTemplateStateSerializer


class WorkflowTransitionFieldSerializer(
    serializers.HyperlinkedModelSerializer
):
    url = MultiKwargHyperlinkedIdentityField(
        label=_('URL'), view_kwargs=(
            {
                'lookup_field': 'transition.workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'transition_id',
                'lookup_url_kwarg': 'workflow_template_transition_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'workflow_template_transition_field_id'
            }
        ), view_name='rest_api:workflow-template-transition-field-detail'
    )
    workflow_template_url = MultiKwargHyperlinkedIdentityField(
        label=_('Workflow template URL'), view_kwargs=(
            {
                'lookup_field': 'transition__workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
        ), view_name='rest_api:workflow-template-detail'
    )
    workflow_transition_id = serializers.IntegerField(
        label=_('Workflow transition ID'), read_only=True,
        source='transition_id'
    )
    workflow_transition_url = MultiKwargHyperlinkedIdentityField(
        label=_('Workflow transition URL'), view_kwargs=(
            {
                'lookup_field': 'transition.workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'transition_id',
                'lookup_url_kwarg': 'workflow_template_transition_id'
            }
        ), view_name='rest_api:workflow-template-transition-detail'
    )

    class Meta:
        fields = (
            'field_type', 'name', 'help_text', 'id', 'label', 'required',
            'url', 'widget', 'widget_kwargs', 'workflow_template_url',
            'workflow_transition_id', 'workflow_transition_url'
        )
        model = WorkflowTransitionField
        read_only_fields = (
            'id', 'url', 'workflow_template_url', 'workflow_transition_id',
            'workflow_transition_url'
        )


class WorkflowTemplateTransitionSerializer(
    serializers.HyperlinkedModelSerializer
):
    destination_state = WorkflowTemplateStateSerializer(
        label=_('Destination state'), read_only=True
    )
    destination_state_id = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the destination state to be added.'
        ), label=_('Destination state ID'),
        source_queryset_method='get_workflow_template_state_queryset',
        write_only=True
    )
    field_list_url = serializers.SerializerMethodField(
        label=_('Field list URL')
    )
    origin_state = WorkflowTemplateStateSerializer(
        label=_('Origin state'), read_only=True
    )
    origin_state_id = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the origin state to be added.'
        ), label=_('Origin state ID'),
        source_queryset_method='get_workflow_template_state_queryset',
        write_only=True
    )
    trigger_list_url = serializers.SerializerMethodField(
        label=_('Trigger list URL')
    )
    url = serializers.SerializerMethodField(
        label=_('URL')
    )
    workflow_template_id = serializers.IntegerField(
        label=_('Workflow template ID'), read_only=True, source='workflow_id'
    )
    workflow_template_url = serializers.SerializerMethodField(
        label=_('Workflow template URL')
    )

    class Meta:
        fields = (
            'condition', 'destination_state', 'destination_state_id',
            'field_list_url', 'id', 'label', 'origin_state',
            'origin_state_id', 'trigger_list_url', 'url',
            'workflow_template_id', 'workflow_template_url'
        )
        model = WorkflowTransition
        read_only_fields = (
            'field_list_url', 'id', 'trigger_list_url', 'url',
            'workflow_template_id', 'workflow_template_url'
        )

    def create(self, validated_data):
        validated_data['destination_state'] = validated_data.pop(
            'destination_state_id'
        )
        validated_data['origin_state'] = validated_data.pop(
            'origin_state_id'
        )

        return super().create(
            validated_data=validated_data
        )

    def get_field_list_url(self, instance):
        return reverse(
            viewname='rest_api:workflow-template-transition-field-list',
            kwargs={
                'workflow_template_id': instance.workflow_id,
                'workflow_template_transition_id': instance.pk
            }, request=self.context['request'], format=self.context['format']
        )

    def get_trigger_list_url(self, instance):
        return reverse(
            viewname='rest_api:workflow-template-transition-trigger-list',
            kwargs={
                'workflow_template_id': instance.workflow_id,
                'workflow_template_transition_id': instance.pk
            }, request=self.context['request'], format=self.context['format']
        )

    def get_url(self, instance):
        return reverse(
            viewname='rest_api:workflow-template-transition-detail', kwargs={
                'workflow_template_id': instance.workflow.pk,
                'workflow_template_transition_id': instance.pk
            }, request=self.context['request'], format=self.context['format']
        )

    def get_workflow_template_state_queryset(self):
        return self.context['workflow_template'].states.all()

    def get_workflow_template_url(self, instance):
        return reverse(
            viewname='rest_api:workflow-template-detail', kwargs={
                'workflow_template_id': instance.workflow.pk
            }, request=self.context['request'], format=self.context['format']
        )

    def update(self, instance, validated_data):
        validated_data['destination_state'] = validated_data.pop(
            'destination_state_id'
        )
        validated_data['origin_state'] = validated_data.pop(
            'origin_state_id'
        )

        return super().update(
            instance=instance, validated_data=validated_data
        )


class WorkflowTemplateTransitionTriggerSerializer(
    serializers.HyperlinkedModelSerializer
):
    event_type = EventTypeSerializer(
        label=_('Event type'), read_only=True
    )
    event_type_id = serializers.CharField(
        label=_('Event Type ID'), source='event_type.event_type.id',
        write_only=True
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_('URL'), view_kwargs=(
            {
                'lookup_field': 'transition.workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'transition_id',
                'lookup_url_kwarg': 'workflow_template_transition_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'workflow_template_transition_trigger_id'
            }
        ), view_name='rest_api:workflow-template-transition-trigger-detail'
    )
    workflow_template_url = MultiKwargHyperlinkedIdentityField(
        label=_('Workflow template URL'), view_kwargs=(
            {
                'lookup_field': 'transition__workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
        ), view_name='rest_api:workflow-template-detail'
    )
    workflow_transition_id = serializers.IntegerField(
        label=_('Workflow transition ID'), read_only=True,
        source='transition_id'
    )
    workflow_transition_url = MultiKwargHyperlinkedIdentityField(
        label=_('Workflow transition URL'), view_kwargs=(
            {
                'lookup_field': 'transition.workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'transition_id',
                'lookup_url_kwarg': 'workflow_template_transition_id'
            }
        ), view_name='rest_api:workflow-template-transition-detail'
    )

    class Meta:
        fields = (
            'event_type', 'event_type_id', 'id', 'url',
            'workflow_template_url', 'workflow_transition_id',
            'workflow_transition_url'
        )
        model = WorkflowTransitionTriggerEvent
        read_only_fields = (
            'id', 'url', 'workflow_template_url', 'workflow_transition_id',
            'workflow_transition_url'
        )

    def create(self, validated_data):
        # Unroll nested source "event_type.event_type.id".
        event_type = validated_data.pop('event_type', None)
        event_type = event_type.get(
            'event_type', {}
        )
        event_type_id = event_type.get('id')

        if event_type_id:
            validated_data['event_type'] = EventType.get(
                id=event_type_id
            ).get_stored_event_type()

        return super().create(validated_data=validated_data)

    def update(self, instance, validated_data):
        # Unroll nested source "event_type.event_type.id".
        event_type = validated_data.pop('event_type', None)
        event_type = event_type.get(
            'event_type', {}
        )
        event_type_id = event_type.get('id')

        if event_type_id:
            validated_data['event_type'] = EventType.get(
                id=event_type_id
            ).get_stored_event_type()

        return super().update(
            instance=instance, validated_data=validated_data
        )

    def validate_event_type_id(self, data):
        try:
            EventType.get(id=data)
        except KeyError:
            raise ValidationError(
                message=_('Unknown or invalid event type ID `%s`') % data
            )
        else:
            return data



from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy as _

from mayan.apps.user_management.permissions import (
    permission_user_edit, permission_user_view
)
from mayan.apps.user_management.querysets import get_user_queryset
from mayan.apps.user_management.views.view_mixins import DynamicExternalUserViewMixin
from mayan.apps.views.generics import (
    SingleObjectCreateView, SingleObjectDeleteView, SingleObjectDetailView,
    SingleObjectEditView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from .forms import ThemeForm, UserThemeSettingForm, UserThemeSettingForm_view
from .icons import (
    icon_theme_create, icon_theme_delete, icon_theme_edit, icon_theme_list,
    icon_theme_setup, icon_user_theme_settings_detail,
    icon_user_theme_settings_edit
)
from .links import link_theme_create
from .models import Theme
from .permissions import (
    permission_theme_create, permission_theme_delete, permission_theme_edit,
    permission_theme_view
)


class ThemeCreateView(SingleObjectCreateView):
    extra_context = {
        'title': _('Create new theme')
    }
    form_class = ThemeForm
    post_action_redirect = reverse_lazy(
        viewname='appearance:theme_list'
    )
    view_permission = permission_theme_create
    view_icon = icon_theme_create

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class ThemeDeleteView(SingleObjectDeleteView):
    model = Theme
    object_permission = permission_theme_delete
    pk_url_kwarg = 'theme_id'
    post_action_redirect = reverse_lazy(
        viewname='appearance:theme_list'
    )
    view_icon = icon_theme_delete

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _('Delete theme: %s') % self.object
        }


class ThemeEditView(SingleObjectEditView):
    form_class = ThemeForm
    model = Theme
    object_permission = permission_theme_edit
    pk_url_kwarg = 'theme_id'
    post_action_redirect = reverse_lazy(
        viewname='appearance:theme_list'
    )
    view_icon = icon_theme_edit

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _('Edit theme: %s') % self.object
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class ThemeListView(SingleObjectListView):
    model = Theme
    object_permission = permission_theme_view
    view_icon = icon_theme_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_theme_setup,
            'no_results_main_link': link_theme_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Themes allow changing the visual appearance without '
                'requiring code changes.'
            ),
            'no_results_title': _(
                'There are no themes'
            ),
            'title': _('Themes')
        }


class UserThemeSettingsDetailsView(
    DynamicExternalUserViewMixin, ExternalObjectViewMixin,
    SingleObjectDetailView
):
    form_class = UserThemeSettingForm_view
    external_object_permission = permission_user_view
    external_object_pk_url_kwarg = 'user_id'
    view_icon = icon_user_theme_settings_detail

    def get_external_object_queryset(self):
        return get_user_queryset(user=self.request.user)

    def get_extra_context(self, **kwargs):
        return {
            'form': UserThemeSettingForm_view(
                instance=self.external_object.theme_settings
            ),
            'object': self.external_object,
            'read_only': True,
            'title': _('Theme settings for user: %s') % self.external_object
        }

    def get_object(self):
        return self.external_object.theme_settings


class UserThemeSettingsEditView(
    DynamicExternalUserViewMixin, ExternalObjectViewMixin,
    SingleObjectEditView
):
    form_class = UserThemeSettingForm
    external_object_permission = permission_user_edit
    external_object_pk_url_kwarg = 'user_id'
    view_icon = icon_user_theme_settings_edit

    def get_external_object_queryset(self):
        return get_user_queryset(user=self.request.user)

    def get_extra_context(self):
        return {
            'title': _(
                'Edit theme settings for user: %s'
            ) % self.external_object,
            'object': self.external_object
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_object(self):
        return self.external_object.theme_settings



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('converter', '0015_auto_20190814_0014')
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='transformation',
            unique_together=set(),
        ),
        migrations.RemoveField(
            model_name='transformation',
            name='content_type',
        ),
        migrations.DeleteModel(
            name='Transformation',
        )
    ]



from django.contrib import messages
from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.views.generics import (
    AddRemoveView, ConfirmView, MultipleObjectDeleteView,
    MultipleObjectFormActionView, SingleObjectCreateView,
    SingleObjectDetailView, SingleObjectEditView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..forms.workflow_template_forms import (
    WorkflowTemplateForm, WorkflowTemplatePreviewForm,
    WorkflowTemplateSelectionForm
)
from ..icons import (
    icon_document_type_workflow_template_list,
    icon_document_workflow_templates_launch, icon_tool_launch_workflows,
    icon_workflow_template_create, icon_workflow_template_delete,
    icon_workflow_template_document_type_list, icon_workflow_template_edit,
    icon_workflow_template_launch, icon_workflow_template_list,
    icon_workflow_template_preview
)
from ..links import link_workflow_template_create
from ..models import Workflow
from ..permissions import (
    permission_workflow_template_create, permission_workflow_template_delete,
    permission_workflow_template_edit, permission_workflow_template_view,
    permission_workflow_tools
)
from ..tasks import (
    task_launch_all_workflows, task_launch_workflow, task_launch_workflow_for
)


class DocumentTypeWorkflowTemplateAddRemoveView(AddRemoveView):
    list_added_title = _('Workflows assigned this document type')
    list_available_title = _('Available workflows')
    main_object_method_add_name = 'workflow_templates_add'
    main_object_method_remove_name = 'workflow_templates_remove'
    main_object_model = DocumentType
    main_object_permission = permission_document_type_edit
    main_object_pk_url_kwarg = 'document_type_id'
    related_field = 'workflows'
    secondary_object_model = Workflow
    secondary_object_permission = permission_workflow_template_edit
    view_icon = icon_document_type_workflow_template_list

    def get_actions_extra_kwargs(self):
        return {'user': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'subtitle': _(
                'Removing a workflow from a document type will also '
                'remove all running instances of that workflow.'
            ),
            'title': _(
                'Workflows assigned the document type: %s'
            ) % self.main_object
        }


class DocumentWorkflowTemplatesLaunchView(MultipleObjectFormActionView):
    error_message = _(
        'Error launching workflows for document "%(instance)s"; '
        '%(exception)s'
    )
    form_class = WorkflowTemplateSelectionForm
    object_permission = permission_workflow_tools
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid.all()
    success_message_plural = _(
        'Workflows launched successfully for %(count)d documents.'
    )
    success_message_single = _(
        'Workflows launched successfully for document "%(object)s".'
    )
    success_message_singular = _(
        'Workflows launched successfully for %(count)d document.'
    )
    title_plural = _(
        'Launch workflows for the selected %(count)d documents.'
    )
    title_single = _('Launch workflow: %(object)s.')
    title_singular = _(
        'Launch workflows for the selected %(count)d document.'
    )
    view_icon = icon_document_workflow_templates_launch

    def get_extra_context(self):
        return {
            'subtitle': _(
                'Workflows already launched or workflows not applicable to '
                'some documents when multiple documents are selected, '
                'will be silently ignored.'
            )
        }

    def get_form_extra_kwargs(self):
        workflows_union = Workflow.objects.filter(
            document_types__in=self.object_list.values('document_type')
        ).distinct()

        result = {
            'help_text': _('Workflows to be launched.'),
            'permission': permission_workflow_tools,
            'queryset': workflows_union,
            'user': self.request.user
        }

        return result

    def object_action(self, form, instance):
        queryset_workflows = AccessControlList.objects.restrict_queryset(
            permission=permission_workflow_tools,
            queryset=form.cleaned_data['workflows'], user=self.request.user
        )

        for workflow in queryset_workflows:
            task_launch_workflow_for.apply_async(
                kwargs={
                    'document_id': instance.pk,
                    'user_id': self.request.user.pk,
                    'workflow_id': workflow.pk
                }
            )


class WorkflowTemplateCreateView(SingleObjectCreateView):
    extra_context = {
        'title': _('Create workflow')
    }
    form_class = WorkflowTemplateForm
    model = Workflow
    post_action_redirect = reverse_lazy(
        viewname='document_states:workflow_template_list'
    )
    view_icon = icon_workflow_template_create
    view_permission = permission_workflow_template_create

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class WorkflowTemplateDeleteView(MultipleObjectDeleteView):
    error_message = _(
        'Error deleting workflow "%(instance)s"; %(exception)s'
    )
    model = Workflow
    object_permission = permission_workflow_template_delete
    pk_url_kwarg = 'workflow_template_id'
    post_action_redirect = reverse_lazy(
        viewname='document_states:workflow_template_list'
    )
    success_message_plural = _('%(count)d workflows deleted successfully.')
    success_message_single = _('Workflow "%(object)s" deleted successfully.')
    success_message_singular = _('%(count)d workflow deleted successfully.')
    title_plural = _('Delete the %(count)d selected workflows.')
    title_single = _('Delete workflow: %(object)s.')
    title_singular = _('Delete the %(count)d selected workflow.')
    view_icon = icon_workflow_template_delete

    def get_extra_context(self):
        return {
            'message': _('All workflow instances will also be deleted.')
        }

    def object_action(self, instance, form=None):
        instance.delete()


class WorkflowTemplateEditView(SingleObjectEditView):
    form_class = WorkflowTemplateForm
    model = Workflow
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_id'
    post_action_redirect = reverse_lazy(
        viewname='document_states:workflow_template_list'
    )
    view_icon = icon_workflow_template_edit

    def get_extra_context(self):
        return {
            'title': _(
                'Edit workflow: %s'
            ) % self.object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class WorkflowTemplateDocumentTypeAddRemoveView(AddRemoveView):
    list_added_title = _('Document types assigned this workflow')
    list_available_title = _('Available document types')
    main_object_method_add_name = 'document_types_add'
    main_object_method_remove_name = 'document_types_remove'
    main_object_model = Workflow
    main_object_permission = permission_workflow_template_edit
    main_object_pk_url_kwarg = 'workflow_template_id'
    related_field = 'document_types'
    secondary_object_model = DocumentType
    secondary_object_permission = permission_document_type_edit
    view_icon = icon_workflow_template_document_type_list

    def get_actions_extra_kwargs(self):
        return {'user': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'subtitle': _(
                'Removing a document type from a workflow will also '
                'remove all running instances of that workflow for '
                'documents of the document type just removed.'
            ),
            'title': _(
                'Document types assigned the workflow: %s'
            ) % self.main_object
        }


class WorkflowTemplateLaunchView(ExternalObjectViewMixin, ConfirmView):
    external_object_class = Workflow
    external_object_permission = permission_workflow_tools
    external_object_pk_url_kwarg = 'workflow_template_id'
    view_icon = icon_workflow_template_launch

    def get_extra_context(self):
        return {
            'subtitle': _(
                'This will launch the workflow for documents that have '
                'already been uploaded.'
            ),
            'title': _('Launch workflow?')
        }

    def view_action(self):
        task_launch_workflow.apply_async(
            kwargs={
                'user_id': self.request.user.pk,
                'workflow_id': self.external_object.pk
            }
        )
        messages.success(
            message=_('Workflow launch queued successfully.'),
            request=self.request
        )


class WorkflowTemplateListView(SingleObjectListView):
    model = Workflow
    object_permission = permission_workflow_template_view
    view_icon = icon_workflow_template_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_workflow_template_list,
            'no_results_main_link': link_workflow_template_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Workflows store a series of states and keep track of the '
                'current state of a document. Transitions are used to change the '
                'current state to a new one.'
            ),
            'no_results_title': _(
                'No workflows have been defined'
            ),
            'title': _('Workflows')
        }


class WorkflowTemplatePreviewView(SingleObjectDetailView):
    form_class = WorkflowTemplatePreviewForm
    model = Workflow
    object_permission = permission_workflow_template_view
    pk_url_kwarg = 'workflow_template_id'
    view_icon = icon_workflow_template_preview

    def get_extra_context(self):
        return {
            'hide_labels': True,
            'object': self.object,
            'title': _('Preview of: %s') % self.object
        }


class ToolLaunchWorkflows(ConfirmView):
    extra_context = {
        'subtitle': _(
            'This will launch all workflows created after documents have '
            'already been uploaded.'
        ),
        'title': _('Launch all workflows?')
    }
    view_icon = icon_tool_launch_workflows
    view_permission = permission_workflow_tools

    def view_action(self):
        task_launch_all_workflows.apply_async(
            kwargs={'user_id': self.request.user.pk}
        )
        messages.success(
            message=_('Workflow launch queued successfully.'),
            request=self.request
        )



from django import forms
from django.forms.formsets import formset_factory
from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.models.document_models import Document
from mayan.apps.templating.fields import ModelTemplateField
from mayan.apps.views.forms import FilteredSelectionForm, Form

from .literals import RELATIONSHIP_CHOICES
from .models.index_template_models import IndexTemplate, IndexTemplateNode
from .permissions import permission_index_template_rebuild


class IndexTemplateEventTriggerRelationshipForm(Form):
    stored_event_type_id = forms.IntegerField(
        widget=forms.widgets.HiddenInput()
    )
    namespace = forms.CharField(
        label=_('Namespace'), required=False, widget=forms.TextInput(
            attrs={'readonly': 'readonly'}
        )
    )
    namespace = forms.CharField(
        label=_('Namespace'), required=False, widget=forms.TextInput(
            attrs={'readonly': 'readonly'}
        )
    )
    label = forms.CharField(
        label=_('Label'), required=False, widget=forms.TextInput(
            attrs={'readonly': 'readonly'}
        )
    )
    relationship = forms.ChoiceField(
        choices=RELATIONSHIP_CHOICES, label=_('Enabled'),
        widget=forms.RadioSelect()
    )


IndexTemplateEventTriggerRelationshipFormSet = formset_factory(
    form=IndexTemplateEventTriggerRelationshipForm, extra=0
)


class IndexTemplateFilteredForm(FilteredSelectionForm):
    class Meta:
        allow_multiple = True
        field_name = 'index_templates'
        help_text = _('Index templates to be queued for rebuilding.')
        label = _('Index templates')
        queryset = IndexTemplate.objects.filter(enabled=True)
        permission = permission_index_template_rebuild
        widget_attributes = {'class': 'select2'}


class IndexTemplateNodeForm(forms.ModelForm):
    """
    A standard model form to allow users to create a new index template node
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['index'].widget = forms.widgets.HiddenInput()
        self.fields['parent'].widget = forms.widgets.HiddenInput()
        self.fields['expression'] = ModelTemplateField(
            label=_('Template'), model=Document,
            model_variable='document', required=False
        )

    class Meta:
        fields = (
            'parent', 'index', 'expression', 'enabled', 'link_documents'
        )
        model = IndexTemplateNode



# This constant may need tweaking as document upload code path changes.
# The value is targeted at making the document upload process fail exactly
# during the MIME type detection phase. Different architectures may need
# different values.
MAXIMUM_HEAP_MEMORY = 140000000



from django.utils.translation import ugettext_lazy as _

from mayan.apps.rest_api import serializers

from .models import Announcement


class AnnouncementSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        extra_kwargs = {
            'url': {
                'label': _('URL'),
                'lookup_url_kwarg': 'announcement_id',
                'view_name': 'rest_api:announcement-detail'
            }
        }
        fields = (
            'end_datetime', 'enabled', 'label', 'id', 'start_datetime',
            'text', 'url'
        )
        model = Announcement



from django.conf.urls import include, url

from .api_views import (
    APIRoot, APIVersionRoot, BatchRequestAPIView, BrowseableObtainAuthToken,
    ProjectInformationAPIView, schema_view
)
from .literals import API_VERSION

api_version_urls = [
    url(
        regex=r'^$', name='api_version_root', view=APIVersionRoot.as_view()
    ),
    url(
        regex=r'^auth/token/obtain/$', name='auth_token_obtain',
        view=BrowseableObtainAuthToken.as_view()
    ),
    url(
        regex=r'^project/$', name='project_information',
        view=ProjectInformationAPIView.as_view()
    ),
    url(
        regex=r'^batch_requests/$', name='batchrequest-create',
        view=BatchRequestAPIView.as_view()
    )
]

api_urls = [
    url(
        regex=r'^swagger(?P<format>.json|.yaml)$', name='schema-json',
        view=schema_view.without_ui(cache_timeout=None),
    ),
    url(
        regex=r'^v{}/'.format(API_VERSION), view=include(api_version_urls)
    ),
    url(
        regex=r'^$', name='api_root', view=APIRoot.as_view()
    )
]

urlpatterns = [
    url(
        regex=r'^swagger/ui/$', name='schema-swagger-ui',
        view=schema_view.with_ui('swagger', cache_timeout=None)
    ),
    url(
        regex=r'^redoc/ui/$', name='schema-redoc',
        view=schema_view.with_ui('redoc', cache_timeout=None)
    ),
    url(
        regex=r'^', view=include(api_urls)
    )
]



from django.conf.urls import url

from .views import (
    DocumentFileDriverAttributeListView, DocumentFileDriverListView,
    DocumentFileSubmitView, DocumentTypeFileMetadataSettingsEditView,
    DocumentTypeFileMetadataSubmitView
)

urlpatterns = [
    url(
        regex=r'^documents/files/(?P<document_file_id>\d+)/drivers/$',
        name='document_file_metadata_driver_list',
        view=DocumentFileDriverListView.as_view()
    ),
    url(
        regex=r'^documents/files/(?P<document_file_id>\d+)/submit/$',
        name='document_file_metadata_single_submit',
        view=DocumentFileSubmitView.as_view()
    ),
    url(
        regex=r'^documents/files/multiple/submit/$',
        name='document_file_metadata_multiple_submit',
        view=DocumentFileSubmitView.as_view()
    ),
    url(
        regex=r'^documents/files/drivers/(?P<document_file_driver_id>\d+)/attributes/$',
        name='document_file_metadata_driver_attribute_list',
        view=DocumentFileDriverAttributeListView.as_view()
    ),
    url(
        regex=r'^document_types/(?P<document_type_id>\d+)/file_metadata/settings/$',
        name='document_type_file_metadata_settings',
        view=DocumentTypeFileMetadataSettingsEditView.as_view()
    ),
    url(
        regex=r'^document_types/submit/$',
        name='document_type_file_metadata_submit',
        view=DocumentTypeFileMetadataSubmitView.as_view()
    )
]



from rest_framework import serializers

from .models import StoredCredential


class StoredCredentialSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        extra_kwargs = {
            'url': {
                'lookup_url_kwarg': 'credential_id',
                'view_name': 'rest_api:credential-detail'
            }
        }
        fields = (
            'label', 'backend_path', 'backend_data', 'id', 'url'
        )
        model = StoredCredential
